#
# PAOFLOW
#
# Copyright 2016-2022 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

#######################################################################
# Fit UPF radial pseudowavefunctions with gaussian orbitals
# Authored: Davide Ceresoli - May 2016
# Revised:  Frank Cerasoli - June 2022
#
# Notes:
# - UPFv1 files must be embedded in <UPF version="1.0">...</UPF> element
# - contraction coefficients for d and f orbitals correspond to the
#   cubic harmonics
#######################################################################

import sys
import numpy as np

spn_map = {'H':1,'He':2,'Li':3,'Be':4,'B':5,'C':6,'N':7,'O':8,'F':9,
           'Ne':10,'Na':11,'Mg':12,'Al':13,'Si':14,'P':15,'S':16,
           'Cl':17,'Ar':18,'K':19,'Ca':20,'Sc':21,'Ti':22,'V':23,
           'Cr':24,'Mn':25,'Fe':26,'Co':27,'Ni':28,'Cu':29,'Zn':30,
           'Ga':31,'Ge':32,'As':33,'Se':34,'Br':35,'Kr':36,'Rb':37,
           'Sr':38,'Y':39,'Zr':40,'Nb':41,'Mo':42,'Tc':43,'Ru':44,
           'Rh':45,'Pd':46,'Ag':47,'Cd':48,'In':49,'Sn':50,'Sb':51,
           'Te':52,'I':53,'Xe':54,'Cs':55,'Ba':56,'La':57,'Ce':58,
           'Pr':59,'Nd':60,'Pm':61,'Sm':62,'Eu':63,'Gd':64,'Tb':65,
           'Dy':66,'Ho':67,'Er':68,'Tm':69,'Yb':70,'Lu':71,'Hf':72,
           'Ta':73,'W':74,'Re':75,'Os':76,'Ir':77,'Pt':78,'Au':79,
           'Hg':80,'Tl':81,'Pb':82,'Bi':83,'Po':84,'At':85,'Rn':86,
           'Fr':87,'Ra':88,'Ac':89,'Th':90,'Pa':91,'U':92,'Np':93,
           'Pu':94,'Am':95,'Cm':96,'Bk':97,'Cf':98,'Es':99,'Fm':100,
           'Md':101,'No':102,'Lr':103,'Rf':104,'Db':105,'Sg':106,
           'Bh':107,'Hs':108,'Mt':109,'Ds':110,'Rg':111,'Cn':112}

def get_atom_no ( n ):
  if n not in spn_map:
    raise Exception(f'Invalid atomic symbol: {n}')
  return spn_map[n]

# Double factorial (n!!)
def fact2 ( n ):
  if n <= 1:
    return 1
  return n * fact2(n-2)


#======================================================================
# GTO orbital
#======================================================================
def gto ( r, l, params ):

  alpha,beta = params[:2]
  coeffs = params[2:]

  gto = np.zeros_like(r)
  for j,c in enumerate(coeffs):
    zeta = alpha / beta**j
    i = np.where(zeta*r**2 > -12)
    gto[i] += c * r[i]**l * np.exp(-zeta*r[i]**2)

  return gto


#======================================================================
# Target function whose least square has to be minimized
#======================================================================
def target ( params, r, rab, wfc, l ):
  return wfc - r*gto(r, l, params)

def target_squared ( params, r, rab, wfc, l ):
  return np.sum(target(params,r,rab,wfc,l)**2)


#======================================================================
# Fit radial wfc with gaussians
#======================================================================
def fit ( nzeta, label, l, r, rab, wfc, threshold, least_squares=True ):

  if len(wfc) != len(r):
    raise Exception('wfc and r have different dimensions.')

  wfc,r = np.array(wfc), np.array(r)

  # Initial alpha and beta
  params0 = np.array([4.,4.] + [1.]*nzeta)

  # Least squares
  if least_squares:
    from scipy.optimize import leastsq

    params,fc,info,msg,ier = leastsq(target, params0, full_output=1,
                                     args=(r,rab,wfc,l), maxfev=50000,
                                     ftol=1e-10, xtol=1e-10)
    print(f'INFO: ier={ier} - mesg={msg}')
    print('INFO: nfev={}'.format(info['nfev']))
    print('INFO: |fvec|^2={}'.format((info['fvec']**2).sum()))
    if ier not in [1, 2, 3, 4]:
      print('ERROR: Gaussion could not be fit to this state.')

  # Minimize
  else:
    from scipy.optimize import minimize

    opt = minimize(target_squared, params0, args=(r,rab,wfc,l),
                   method='CG', tol=1e-10)
    params = opt.x
    if not opt.success:
      print('ERROR: opt.status={}'.format(opt.status))
      print('ERROR: opt.message={}'.format(opt.message))
      print('ERROR: opt.nfev={}'.format(opt.nfev))
      print('ERROR: opt.fun={}'.format(opt.fun))

  alpha,beta = params[:2]
  print(f'alpha = {alpha}, beta = {beta}')

  expon = []
  coeffs = np.sqrt(fact2(2*l+1)/(4*np.pi)) * params[2:]
  for j,c in enumerate(coeffs):
    zeta = alpha / beta**j
    expon.append(zeta)
    print(f'coeff = {c}, zeta = {zeta}')

  res = target_squared(params, r, rab, wfc, l)
  print(f'Fit result: {res}')

  exit_code = 0 if np.abs(res) <= threshold else 1

  return coeffs, expon, exit_code


#======================================================================
# Construct basis string for orbitals and write it to file
#======================================================================
def build_basis_dict ( fname, labels, ls, coefficients, exponents ):

  basis = []

  for il,label in enumerate(labels):

    l = ls[il]
    expon = exponents[il]
    coeffs = coefficients[il]

    lbasis = []
    nzeta = len(coeffs)

    if l == 0:
      ibasis = []
      for i,c in enumerate(coeffs):
        ibasis.append((0, 0, 0, c, expon[i]))
      lbasis.append(ibasis)

    elif l == 1:
      for n in range(3):
        ibasis = []
        lind = [0]*3
        lind[2-n] = 1
        for i,c in enumerate(coeffs):
          ibasis.append((*lind, c, expon[i]))
        lbasis.append(ibasis)

    elif l == 2:

      # 1/(2*sqrt(3))*(2*z2 - x2 - y2)
      ibasis = []
      for n in range(3):
        # ibasis = []
        lind = [0]*3
        lind[2-n] = 2
        fact = (1 if n==0 else -0.5) / np.sqrt(3)
        for i,c in enumerate(coeffs):
          ibasis.append((*lind, fact*c, expon[i]))
      lbasis.append(ibasis)

      # xz
      lbasis.append([(1,0,1,c,expon[i]) for i,c in enumerate(coeffs)])

      # yz
      lbasis.append([(0,1,1,c,expon[i]) for i,c in enumerate(coeffs)])

      # 1/2 * (x2 - y2)
      ibasis = []
      for i,c in enumerate(coeffs):
        ibasis.append((2,0,0,0.5*c,expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((0,2,0,-0.5*c,expon[i]))
      lbasis.append(ibasis)

      # xy
      lbasis.append([(1,1,0,c,expon[i]) for i,c in enumerate(coeffs)])

    elif l == 3:
      # fz3, fxz2, fyz2, fz(x2-y2), fxyz, fx(x3-3y2), fy(3x2-y2)

      # 1/(2*sqrt(15)) * z*(2*z2 - 3*x2 - 3*y2)
      fact = 0.5 / np.sqrt(15)
      ibasis = []
      for i,c in enumerate(coeffs):
        ibasis.append((0,0,3,2*fact*c,expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((2,0,1,-3*fact*c,expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((0,2,1,-3*fact*c,expon[i]))
      lbasis.append(ibasis)

      # 1/(2*sqrt(10)) * x*(4*z2 - x2 - y2)
      ibasis = []
      fact = 0.5 / np.sqrt(10)
      for i,c in enumerate(coeffs):
        ibasis.append((1, 0, 2, 4*fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((0, 0, 3, -fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((1, 2, 0, -fact*c, expon[i]))
      lbasis.append(ibasis)

      # 1/(2*sqrt(10)) * y*(4*z2 - x2 - y2)
      ibasis = []
      fact = 0.5 / np.sqrt(10)
      for i,c in enumerate(coeffs):
        ibasis.append((0, 1, 2, 4*fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((2, 1, 0, -fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((0, 3, 0, -fact*c, expon[i]))
      lbasis.append(ibasis)

      # 1/2 * z*(x2 - y2)
      ibasis = []
      fact = 0.5
      for i,c in enumerate(coeffs):
        ibasis.append((2, 0, 1, fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((0, 2, 1, -fact*c, expon[i]))
      lbasis.append(ibasis)

      # x*y*z
      for i,c in enumerate(coeffs):
        lbasis.append([(1, 1, 1, c, expon[i]) for i,c in enumerate(coeffs)])

      # 1/(2*sqrt(6)) * x*(x2 - 3*y2)
      ibasis = []
      fact = 0.5 / np.sqrt(6)
      for i,c in enumerate(coeffs):
        ibasis.append((3, 0, 0, fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((1, 2, 0, -3*fact*c, expon[i]))
      lbasis.append(ibasis)

      # 1/(2*sqrt(6)) * y*(3*x2 - y2)
      ibasis = []
      fact = 0.5 / np.sqrt(6)
      for i,c in enumerate(coeffs):
        ibasis.append((2, 1, 0, 3*fact*c, expon[i]))
      for i,c in enumerate(coeffs):
        ibasis.append((0, 3, 0, -fact*c, expon[i]))
      lbasis.append(ibasis)

    basis.append(lbasis)

  return basis

def write_basis_file ( fname, atom_no, labels, ls, coefficients, exponents ):

  rstr = f'basis_data = {{ {atom_no} : [\n'

  for il,label in enumerate(labels):

    l = ls[il]
    expon = exponents[il]
    coeffs = coefficients[il]

    nzeta = len(coeffs)
    rstr += f'# label= {label} l= {l}\n[[\n'

    fcon = '],  [\n'
    fbuf = '   {},\n'
    fpat = '({},{},{},{:.10f},{:.10f})'
    fline = lambda x,y,z,a,b : fbuf.format(fpat.format(x,y,z,a,b))

    if l == 0:
      for i,c in enumerate(coeffs):
        rstr += fline(0, 0, 0, c, expon[i])

    elif l == 1:
      for n in range(3):
        lind = [0]*3
        lind[2-n] = 1
        for i,c in enumerate(coeffs):
          rstr += fline(*lind, c, expon[i])
        if n < 2:
          rstr += fcon

    elif l == 2:

      # 1/(2*sqrt(3))*(2*z2 - x2 - y2)
      for n in range(3):
        lind = [0]*3
        lind[2-n] = 2
        fact = (1 if n==0 else -0.5) / np.sqrt(3)
        for i,c in enumerate(coeffs):
          rstr += fline(*lind, fact*c, expon[i])
      rstr += fcon

      # xz
      for i,c in enumerate(coeffs):
        rstr += fline(1, 0, 1, c, expon[i])
      rstr += fcon

      # yz
      for i,c in enumerate(coeffs):
        rstr += fline(0, 1, 1, c, expon[i])
      rstr += fcon

      # 1/2 * (x2 - y2)
      for i,c in enumerate(coeffs):
        rstr += fline(2, 0, 0, 0.5*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(0, 2, 0, -0.5*c, expon[i])
      rstr += fcon

      # xy
      for i,c in enumerate(coeffs):
        rstr += fline(1, 1, 0, c, expon[i])

    elif l == 3:
      # fz3, fxz2, fyz2, fz(x2-y2), fxyz, fx(x3-3y2), fy(3x2-y2)

      # 1/(2*sqrt(15)) * z*(2*z2 - 3*x2 - 3*y2)
      fact = 0.5 / np.sqrt(15)
      for i,c in enumerate(coeffs):
        rstr += fline(0, 0, 3, 2*fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(2, 0, 1, -3*fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(0, 2, 1, -3*fact*c, expon[i])
      rstr += fcon

      # 1/(2*sqrt(10)) * x*(4*z2 - x2 - y2)
      fact = 0.5 / np.sqrt(10)
      for i,c in enumerate(coeffs):
        rstr += fline(1, 0, 2, 4*fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(0, 0, 3, -fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(1, 2, 0, -fact*c, expon[i])
      rstr += fcon

      # 1/(2*sqrt(10)) * y*(4*z2 - x2 - y2)
      fact = 0.5 / np.sqrt(10)
      for i,c in enumerate(coeffs):
        rstr += fline(0, 1, 2, 4*fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(2, 1, 0, -fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(0, 3, 0, -fact*c, expon[i])
      rstr += fcon

      # 1/2 * z*(x2 - y2)
      fact = 0.5
      for i,c in enumerate(coeffs):
        rstr += fline(2, 0, 1, fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(0, 2, 1, -fact*c, expon[i])
      rstr += fcon

      # x*y*z
      for i,c in enumerate(coeffs):
        rstr += fline(1, 1, 1, c, expon[i])
      rstr += fcon

      # 1/(2*sqrt(6)) * x*(x2 - 3*y2)
      fact = 0.5 / np.sqrt(6)
      for i,c in enumerate(coeffs):
        rstr += fline(3, 0, 0, fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(1, 2, 0, -3*fact*c, expon[i])
      rstr += fcon

      # 1/(2*sqrt(6)) * y*(3*x2 - y2)
      fact = 0.5 / np.sqrt(6)
      for i,c in enumerate(coeffs):
        rstr += fline(2, 1, 0, 3*fact*c, expon[i])
      for i,c in enumerate(coeffs):
        rstr += fline(0, 3, 0, -fact*c, expon[i])
      rstr += fcon

    rstr += ']],\n'
  rstr = rstr[:-1] + ']}\n'

  with open(fname, 'w') as f:
    f.write(rstr)

  print(f'INFO: File {fname} created.\n')


def read_atom_no_xml ( upf_version, root ):

  ele = None
  text = root.find('PP_HEADER')
  if upf_version == 1:
    text = text.text.split()
    ind = text.index('Element')
    ele = text[ind-1].strip()
  elif upf_version == 2:
    ele = text.attrib['element'].strip()
  else:
    raise Exception('ERROR: Supported UPF version are v1 and v2')

  no = get_atom_no(ele)
  print(f'INFO: element={ele}, atomic number={no}')
  return ele, no

def read_upf ( upf_version, root ):

  ls = []
  wfcs = []
  labels = []
  text = root.find('PP_MESH/PP_R').text
  r = np.array([float(v) for v in text.split()])
  text = root.find('PP_MESH/PP_RAB').text
  rab = np.array([float(v) for v in text.split()])

  if upf_version == 1:
    from io import StringIO

    chi = root.find('PP_PSWFC')
    if chi is None:
      raise Exception('ERROR: Cannot locate PP_PSWFC tag.')

    nlines = r.shape[0]//4
    if r.shape[0] % 4 != 0:
      nlines += 1

    text = StringIO(chi.text)
    line = text.readline()
    while line != '':

      if line == '\n':
        continue

      label,l,occ,_ = line.split()
      l,occ = int(l),float(occ)

      wfc = []
      for _ in range(nlines):
        wfc += list(map(float, text.readline().split()))

      ls.append(l)
      wfcs.append(wfc)
      labels.append(label)

      line = text.readline()

  elif upf_version == 2:
    ind = 1
    fstr = 'PP_PSWFC/PP_CHI.{}'
    chi = root.find(fstr.format(ind))
    while chi is not None:

      label = chi.attrib['label']
      l = int(chi.attrib['l'])
      wfc = [float(v) for v in chi.text.split()]
      if len(wfc) != r.shape[0]:
        msg = 'ERROR: wfc and radial grid have different dimension'
        raise Exception(msg)

      ls.append(l)
      wfcs.append(wfc)
      labels.append(label)

      ind += 1
      chi = root.find(fstr.format(ind))

  else:
    raise Exception('ERROR: Supported UPF version are v1 and v2')

  wfcs = np.array(wfcs)
  for i,w in enumerate(wfcs):
    l = ls[i]
    label = labels[i]
    norm = np.sum(rab * w**2)
    print(f'INFO: Fitting pswfc {label} l={l} norm={norm}')

  return r,rab,labels,ls,wfcs


def gaussian_fit ( xml_file, threshold=0.5 ):
  from xml.etree import ElementTree as ET
  from os.path import dirname,join

  atno = -1
  nzeta = 2
  basis = None
  optimized = False
  while not optimized and nzeta < 6:

    root = None
    try:
      print(f'INFO: Fitting file {xml_file} with {nzeta} gaussians')
      with open(xml_file, 'r') as f:
        xml_content = f.read()
        root = ET.fromstring(xml_content)

      upf_version = int(root.attrib['version'].split('.')[0])
    except Exception as e:
      print(f'ERROR: Could not read the xml file: {xml_file}')
      raise e

    ele,atno = read_atom_no_xml(upf_version, root)
    r,rab,labels,ls,wfcs = read_upf(upf_version, root)

    failed = False
    coeffs,exponents = [],[]
    for i,lab in enumerate(labels):
      coef,expon,exit_code = fit(nzeta, lab, ls[i], r,
                                 rab, wfcs[i], threshold)
      if exit_code == 0:
        coeffs.append(coef)
        exponents.append(expon)
      else:
        failed = True
        break

    if failed:
      nzeta += 1
      continue

    optimized = True
    base_dir = dirname(xml_file)
    fname = join(base_dir, f'{ele}_basis.py')
    basis = build_basis_dict(fname, labels, ls, coeffs, exponents)

  if nzeta >= 6:
    raise Exception('ERROR: Could not optimize the wfcs')

  if atno == -1:
    raise Exception('ERROR: Could not determine atomic information')

  return atno, basis



from numpy import ndarray

def read_band_path_PAO ( fname ):
  '''
  '''
  import numpy as np

  tags = []
  npnts = []
  with open(fname, 'r') as f:
    ls = f.readline().split()
    while not len(ls) == 0:
      tags.append(ls[0])
      npnts.append(int(ls[1]))
      ls = f.readline().split()

  kcnt = 0
  ftags = []
  findex = [0]
  for i in range(len(tags)-1):
    if tags[i] == 'G' or tags[i] == 'gG':
      tags[i] = r'$\Gamma$'

    if npnts[i] == 0:
      ftags[-1] += '|' + tags[i]
    else:
      ftags.append(tags[i])
      findex.append(npnts[i] + findex[-1])
  ftags.append(tags[-1] if not tags[-1]=='G' else r'$\Gamma$')
    
  return findex, ftags


def read_dos_PAO ( fname ):
  '''
  '''
  import numpy as np

  es = []
  dos = []
  with open(fname, 'r') as f:
    for l in f.readlines():
      ls = l.split()
      es.append(float(ls[0]))
      dos.append(float(ls[1]))

  es = np.array(es)
  dos = np.array(dos)

  return es, dos


def read_bands_PAO ( fname ):
  '''
  '''
  import numpy as np

  bands = []
  with open(fname, 'r') as f:
    for l in f.readlines():
      bands.append([float(v) for v in l.split()[1:]])

  return np.array(bands).T


def read_transport_PAO ( fname ):

  import numpy as np

  nene = 0
  ntemp = 0
  enes = temps = tensors = None

  with open(fname, 'r') as f:
    lines = f.readlines()

    nl = len(lines)
    ftemp = float(lines[ntemp].split()[0])
    ptemp = ftemp
    while ftemp == ptemp and nene < nl-1:
      nene += 1
      ptemp = float(lines[nene].split()[0])
    if nene == nl-1:
      nene += 1

    while ntemp*nene < nl:
      ntemp += 1

    enes = np.empty(nene, dtype=float)
    temps = np.empty(ntemp, dtype=float)
    tensors = np.empty((ntemp,nene,3,3), dtype=float)

    iL = 0
    for i in range(ntemp):
      temps[i] = float(lines[iL].split()[0])
      for j in range(nene):
        ls = lines[iL].split()
        if i == 0:
          enes[j] = float(ls[1])
        for k in range(3):
          tensors[i,j,k,k] = float(ls[2+k])
        for ik,k in enumerate([(0,1), (0,2), [1,2]]):
          tensors[i,j,k[0],k[1]] = tensors[i,j,k[1],k[0]] = float(ls[5+ik])
        iL += 1
 
    return enes, temps, tensors




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

report_pre_reqs = 'SUGGESTION: %s must be called before %s'

module_pre_reqs = { 'add_external_fields'      : ['projectability'],\
                    'pao_hamiltonian'          : ['projectability'],\
                    'z2_pack'                  : ['pao_hamiltonian'],\
                    'atomic_orbitals'           : ['pao_hamiltonian'], \
                    'doubling_Hamiltonian'     : ['pao_hamiltonian'],\
                    'cutting_Hamiltonian'      : ['pao_hamiltonian'],\
                    'wave_function_projection' : ['pao_hamiltonian'],\
                    'bands'                    : ['pao_hamiltonian'],\
                    'topology'                 : ['bands', 'spin_operator'],\
                    'interpolated_hamiltonian' : ['pao_hamiltonian'],\
                    'pao_eigh'                 : ['pao_hamiltonian'],\
                    'dos'                      : ['pao_eigh'],\
                    'fermi_surface'            : ['pao_eigh'],\
                    'spin_texture'             : ['pao_eigh'],
                    'gradient_and_momenta'     : ['pao_eigh'],\
                    'ipr'                      : ['pao_eigh'],\
                    'berry_phase'              : ['pao_hamiltonian'],\
                    'real_space_wfc'           : ['gradient_and_momenta', 'atomic_orbitals'], \
                    'adaptive_smearing'        : ['gradient_and_momenta'],\
                    'anomalous_Hall'           : ['gradient_and_momenta'],\
                    'spin_Hall'                : ['gradient_and_momenta', 'spin_operator'],\
                    'transport'                : ['gradient_and_momenta'],\
                    'effective_mass'           : ['gradient_and_momenta'],\
                    'dielectric_tensor'        : ['gradient_and_momenta'] }


key_error_strings = { 'U' : 'Projections must be computed before projectability. Call \'read_atomic_proj_QE\' to read projections from the output of projwfc.x',\
                      'HRs'  : 'HRs must first be calculated with \'build_pao_hamiltonian\'',\
                      'd2Ed2k'  : 'd2Ed2k must first be calculated using band_curvature set to True in \'gradient_and_momenta\'',\
                      'Hks'  : 'Hks must first be calculated with \'build_pao_hamiltonian\''}



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .


def do_transport ( data_controller, temps, ene, velkp, channels, weights, do_hall, write_to_file, save_tensors ):
  import numpy as np
  from os.path import join
  from numpy import linalg as npl
  from .do_Boltz_tensors import do_Boltz_tensors,do_Boltz_tensors_hall

  comm,rank = data_controller.comm,data_controller.rank
  arrays,attr = data_controller.data_dicts()

  esize = ene.size
  snktot = arrays['E_k'].shape[0]
  siemen_conv,temp_conv,hall_SI = 6.9884,11604.52500617,9.248931724005307e-13
  nspin,t_tensor = attr['nspin'],arrays['t_tensor']
  spin_mult = 1. if nspin==2 or attr['dftSO'] else 2.

  for ispin in range(nspin):
    # Quick function opens file in output folder with name 's'
    if write_to_file:
      ojf = lambda st,sp : open(join(attr['opath'],'%s_%d.dat'%(st,sp)), 'w')

      fsigma = ojf('sigma', ispin)
      fPF = ojf('PF', ispin)
      fkappa = ojf('kappa', ispin)
      fSeebeck = ojf('Seebeck', ispin)
      fsigmadk = ojf('sigmadk', ispin) if attr['smearing']!=None else None
      if do_hall:
        fhall = ojf('hall_trace', ispin)

    for iT,temp in enumerate(temps):

      itemp = temp/temp_conv

      # Quick function to write Transport Formatted line to file
      wtup = lambda fn,tu : fn.write('%8.2f % .5f % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e\n'%tu)

      # Quick function to get tuple elements to write
      gtup = lambda tu,i : (temp,ene[i],tu[0,0,i],tu[1,1,i],tu[2,2,i],tu[0,1,i],tu[0,2,i],tu[1,2,i])

      if do_hall:
        wtup_hall = lambda fn,tu : fn.write('%8.2f % .5f % 9.5e \n'%tu)
        gtup_hall = lambda tu,i : (temp,ene[i],tu[i])

      if attr['smearing'] is not None:
        L0,_,_ = do_Boltz_tensors(data_controller, attr['smearing'], itemp, ene, velkp, ispin, channels, weights)
        #----------------------
        # Conductivity (in units of 1.e21/Ohm/m/s)
        #----------------------
        if rank == 0:
          # convert in units of 10*21 siemens m^-1 s^-1
          L0 *= spin_mult*siemen_conv/attr['omega']
          # convert in units of siemens m^-1 s^-1
          sigma = L0*1.e21

          if write_to_file:   
            for i in range(esize):
              wtup(fsigmadk, gtup(sigma,i))
            sigma = None

        comm.Barrier()

      L0,L1,L2 = do_Boltz_tensors(data_controller, None, itemp, ene, velkp, ispin, channels, weights)

      if do_hall: 
        L0_hall = do_Boltz_tensors_hall(data_controller, None, itemp, ene, velkp, ispin, channels, weights)

      if rank == 0:
        #----------------------
        # Conductivity (in units of /Ohm/m/s)
        # convert in units of 10*21 siemens m^-1 s^-1
        #----------------------
        L0_unconverted = L0*spin_mult/attr['omega']
        L0 *= spin_mult*siemen_conv/attr['omega']
        sigma = L0*1.e21 # convert in units of siemens m^-1 s^-1
        if write_to_file:
          for i in range(esize):
            wtup(fsigma, gtup(sigma,i))
          sigma = None
        if save_tensors:
          arrays['sigma'] = sigma

        if do_hall:
          L0_hall *= spin_mult/(attr['omega'])
          R_hall = np.zeros((3,3,3,esize), dtype=float)
          R_hall_trace = np.zeros((esize), dtype=float)
          for n in range(esize):
            try:
              for r in range(3):
                R_hall[:,:,r,n] = npl.inv(L0_unconverted[:,:,n]) @ L0_hall[:,:,r,n] @ npl.inv(L0_unconverted[:,:,n])
                #----------------------   
                # The equivalent to the trace of the Hall tensor is an average
                # over the even permutations of [0, 1, 2].
                #----------------------  
              R_hall_trace[n] = (R_hall[0,1,2,n]+R_hall[2,0,1,n]+R_hall[1,2,0,n])*hall_SI/3

            except Exception as e:
              from .report_exception import report_exception
              print('check t_tensor components - matrix cannot be singular')
              report_exception()
              raise
          if write_to_file:
            for i in range(esize):
              wtup_hall(fhall, gtup_hall(R_hall_trace,i))
          if save_tensors:
            arrays['R_hall_trace'] = R_hall_trace

        #----------------------
        # Seebeck (in units of V/K)
        # convert in units of 10^21 Amperes m^-1 s^-1
        #----------------------
        L1 *= spin_mult*siemen_conv/(temp*attr['omega'])

        S = np.zeros((3,3,esize), dtype=float)

        for n in range(esize):
          try:
            S[:,:,n] = -1.*npl.inv(L0[:,:,n]) @ L1[:,:,n]
          except Exception as e:
            from .report_exception import report_exception
            print('check t_tensor components - matrix cannot be singular')
            report_exception()
            raise
        if write_to_file:
          for i in range(esize):
            wtup(fSeebeck, gtup(S,i))
        if save_tensors:
          arrays['S'] = S

        #----------------------
        # Electron thermal conductivity ((in units of W/m/K/s)
        # convert in units of kg m s^-4
        #----------------------
        L2 *= spin_mult*siemen_conv*1.e15/(temp*attr['omega'])

        kappa = np.zeros((3,3,esize),dtype=float)
        for n in range(esize):
          kappa[:,:,n] = (L2[:,:,n] - temp*L1[:,:,n] @ npl.inv(L0[:,:,n]) @ L1[:,:,n])*1.e6
        L1 = L2 = None
        if write_to_file:
          for i in range(esize):
            wtup(fkappa, gtup(kappa,i))
          kappa = None
        if save_tensors:
          arrays['kappa'] = kappa

        PF = np.zeros((3,3,esize), dtype=float)
        for n in range(esize):
          PF[:,:,n] = np.dot(np.dot(S[:,:,n],L0[:,:,n]),S[:,:,n])*1.e21
        S = L0 = None
        if write_to_file:
          for i in range(esize):
            wtup(fPF, gtup(PF,i))
          PF = None
      comm.Barrier()

    if write_to_file:
      fsigma.close()
      fPF.close()
      fkappa.close()
      fSeebeck.close()
      if attr['smearing'] is not None:
        fsigmadk.close()
      if do_hall:
        fhall.close()

    if save_tensors:
      data_controller.broadcast_single_array('sigma', dtype=float)
      data_controller.broadcast_single_array('S', dtype=float)
      data_controller.broadcast_single_array('kappa', dtype=float)
      if do_hall:
        data_controller.broadcast_single_array('R_hall_trace', dtype=float)



from matplotlib import pyplot as plt

class GPAO:

  def __init__ ( self ):
    pass

  def plot_dos ( self, fname, title=None, x_lim=None, y_lim=None, vertical=False, col='black' ):
    '''
      Plot the density of states.

      Arguments:
        fname (str): File name (including relative path)
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        vertical (bool): Set True to plot energy on the y-axis and elec/eV on the x-axis.
        col (str or tuple): A string recognized by matplotlib or a 3-tuple (R,G,B)
    '''
    from .defs.read_pao_output import read_dos_PAO
    from .graphics.plot_functions import plot_dos

    es, dos = read_dos_PAO(fname)

    plot_dos(es, dos, title, x_lim, y_lim, vertical, col)


  def plot_pdos ( self, fnames, title=None, x_lim=None, y_lim=None, vertical=False, cols=None, labels=None, legend=True, psum_inds=None ):
    '''
      Plot the projected density of states.

      Arguments:
        fnames (list): List of file names (including relative path)
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        vertical (bool): Set True to plot energy on the y-axis and elec/eV on the x-axis.
        cols (str or tuple): A list of string recognized by matplotlib or 3-tuples (R,G,B) with the same dimension as the number of files
        labels (list): List of strings with same dimension as the number of files
        psum_inds (list): A list of lists. Each inner list contains indices for the dos files elements to sum together. There is one line plotted for each inner list.
    '''
    from .defs.read_pao_output import read_dos_PAO
    from .graphics.plot_functions import plot_pdos
    import numpy as np

    es = None
    dos = []
    for fn in fnames:
      es,ds = read_dos_PAO(fn)
      dos.append(ds)

    if psum_inds is not None:
      ndos = np.zeros((len(psum_inds),len(es)), dtype=float)
      for i,p in enumerate(psum_inds):
        for v in p:
          ndos[i,:] += dos[v]
      dos = ndos

    plot_pdos(es, dos, title, x_lim, y_lim, vertical, cols, labels, legend)


  def plot_bands ( self, fname, sym_points=None, title=None, label=None, y_lim=None, col='black' ):
    '''
      Plot the band structure

      Arguments:
        fname (str): File name (including relative path)
        sym_points (str or tuple): File name for the kpath_points produced by PAOFLOW. Otherwise, provide a tuple of two lists. The first contains indices of the high sym points, the second contains labels for the high sym points.
        title (str): A title for the plot
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        col (str or tuple): A string recognized by matplotlib or a 3-tuple (R,G,B)
    '''
    from .graphics.plot_functions import plot_bands
    from .defs.read_pao_output import read_bands_PAO

    if not sym_points is None:
      if type(sym_points) is str:
        from .defs.read_pao_output import read_band_path_PAO
        sym_points = read_band_path_PAO(sym_points)
    plot_bands(read_bands_PAO(fname), sym_points, title, label, y_lim, col)


  def plot_berry ( self, fname, sym_points=None, title=None, label=None, x_lim=None, y_lim=None, col='black', dos_ticks=False ):
    '''
      Plot the band structure

      Arguments:
        fname (str): File name (including relative path)
        sym_points (str or tuple): File name for the kpath_points produced by PAOFLOW. Otherwise, provide a tuple of two lists. The first contains indices of the high sym points, the second contains labels for the high sym points.
        title (str): A title for the plot
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        col (str or tuple): A string recognized by matplotlib or a 3-tuple (R,G,B)
    '''
    from .graphics.plot_functions import plot_bands
    from .defs.read_pao_output import read_dos_PAO
    import numpy as np

    if title is None:
      title = 'Berry curvature vs k-point'
    if label is None:
      label = '$\Omega$($\\bf{k}$)'
    if sym_points is not None:
      if type(sym_points) is str:
        from .defs.read_pao_output import read_band_path_PAO
        sym_points = read_band_path_PAO(sym_points)
    path,omega = read_dos_PAO(fname)
    plot_bands(np.array([omega]), sym_points, title, label, y_lim, col)


  def plot_dos_beside_bands ( self, fn_dos, fn_bands, sym_points=None, title=None, band_label=None, x_lim=None, y_lim=None, col='black', dos_ticks=False ):
    '''
      Plot the density of states beside the band structure

      Arguments:
        fn_dos (str): File name for dos (including relative path)
        fn_bands (str): File name for bands (including relative path)
        sym_points (str or tuple): File name for the kpath_points produced by PAOFLOW. Otherwise, provide a tuple of two lists. The first contains indices of the high sym points, the second contains labels for the high sym points.
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        col (str or tuple): A string recognized by matplotlib or a 3-tuple (R,G,B)
    '''
    from .defs.read_pao_output import read_dos_PAO,read_bands_PAO
    from .graphics.plot_functions import plot_dos_beside_bands

    if not sym_points is None:
      if type(sym_points) is str:
        from .defs.read_pao_output import read_band_path_PAO
        sym_points = read_band_path_PAO(sym_points)

    bands = read_bands_PAO(fn_bands)
    es, dos = read_dos_PAO(fn_dos)

    plot_dos_beside_bands(es, dos, bands, sym_points, band_label, title, x_lim, y_lim, col, dos_ticks)    


  def plot_berry_under_bands ( self, fn_berry, fn_bands, sym_points=None, title=None, x_lim=None, y_lim=None, col='black', dos_ticks=False, band_label=None, berry_label=None ):
    '''
      Plot the berry phase below the band structure

      Arguments:
        fn_berry (str): File name for berry (including relative path)
        fn_bands (str): File name for bands (including relative path)
        sym_points (str or tuple): File name for the kpath_points produced by PAOFLOW. Otherwise, provide a tuple of two lists. The first contains indices of the high sym points, the second contains labels for the high sym points.
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        col (str or tuple): A string recognized by matplotlib or a 3-tuple (R,G,B)
    '''
    from .defs.read_pao_output import read_dos_PAO,read_bands_PAO
    from .graphics.plot_functions import plot_berry_under_bands
    import numpy as np

    if not sym_points is None:
      if type(sym_points) is str:
        from .defs.read_pao_output import read_band_path_PAO
        sym_points = read_band_path_PAO(sym_points)

    bands = read_bands_PAO(fn_bands)
    path,omega = read_dos_PAO(fn_berry)

    plot_berry_under_bands(omega, bands, sym_points, title, band_label, berry_label, x_lim, y_lim, col, dos_ticks)    


  def plot_electrical_conductivity ( self, fname, t_ele=[(0,0),(1,1),(2,2)], vE=None, title='Sigma vs Energy', x_lim=None, y_lim=None, col=[(1,0,0),(0,1,0),(0,0,1)], legend=True ):
    '''
      Plot the electrical conductivity. If multiple Temperatures are computed the default behavior is to plot the full energy range for every temperature. If a conductivity vs temperature plot is desired, set vE to the energy at which conductivity should be collected for each temperature.

      Arguments:
        fname (str): File name (including relative path)
        t_ele (list): Tensor elements as tuple pairs (e.g. (1,2) for (y,z)). Default behavior is to plot the 3 diagonal elements seprately. Providing an empty list will average the diagonal components
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        col (list): A list of 3-tuples (R,G,B), one for each tensor element.
        vE (float): Set to an energy to plot the conductivity vs temperature. The value of conductivity is taken at the provided energy for each temperature.
    '''
    from .defs.read_pao_output import read_transport_PAO
    from .graphics.plot_functions import plot_tensor

    x_label = '$Energy (eV)$'
    y_label = 'Conductivity $(\Omega m s)^{-1}$'
    enes, temps, tensors = read_transport_PAO(fname)
    for i,temp in enumerate(temps):
      ttitle = title + ', T={}'.format(temp)
      plot_tensor(enes, tensors[i], t_ele, ttitle, x_lim, y_lim, x_label, y_label, col, legend, min_zero=True)


  def plot_seebeck ( self, fname, t_ele=[(0,0),(1,1),(2,2)], vE=None, title='Seebeck vs Energy', x_lim=None, y_lim=None, col=[(1,0,0),(0,1,0),(0,0,1)], legend=True ):
    '''
      Plot the Seebeck coefficient. If multiple Temperatures are computed the default behavior is to plot the full energy range for every temperature. If a conductivity vs temperature plot is desired, set vE to the energy at which conductivity should be collected for each temperature.

      Arguments:
        fname (str): File name (including relative path)
        t_ele (list): Tensor elements as tuple pairs (e.g. (1,2) for (y,z)). Default behavior is to plot the 3 diagonal elements seprately. Providing an empty list will average the diagonal components
        vE (float): Set to an energy to plot the conductivity vs temperature. The value of conductivity is taken at the provided energy for each temperature.
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        col (list): A list of 3-tuples (R,G,B), one for each tensor element.
    '''
    from .defs.read_pao_output import read_transport_PAO
    from .graphics.plot_functions import plot_tensor

    x_label = 'Energy (eV)'
    y_label = 'Seebeck ($\mu$V/K)'
    enes, temps, tensors = read_transport_PAO(fname)
    for i,temp in enumerate(temps):
      ttitle = title + ', T={}'.format(temp)
      plot_tensor(enes, tensors[i]*1e6, t_ele, ttitle, x_lim, y_lim, x_label, y_label, col, legend, min_zero=False)


  def plot_shc ( self, fname, title='SHC vs Energy', x_lim=None, y_lim=None, cols=None, legend=True ):
    '''
      Plot the Seebeck coefficient. If multiple Temperatures are computed the default behavior is to plot the full energy range for every temperature. If a conductivity vs temperature plot is desired, set vE to the energy at which conductivity should be collected for each temperature.

      Arguments:
        fname (str): File name (including relative path)
        title (str): A title for the plot
        x_lim (tuple): Pair of axis limits (x_min, x_max)
        y_lim (tuple): Pair of axis limits (y_min, y_max)
        cols (list): A list of 3-tuples (R,G,B), one for each tensor element.
    '''
    from .defs.read_pao_output import read_dos_PAO
    import numpy as np

    x_label = 'Energy (eV)'
    y_label = '$\sigma$ ($\Omega$cm)$^{-1}$'

    if isinstance(fname, str):
      from .graphics.plot_functions import plot_dos
      es,shc = read_dos_PAO(fname)
      if y_lim is None:
        y_lim = 1.1*np.array([np.min(shc),np.max(shc)])
      plot_dos(es, shc, title, x_lim, y_lim, False, cols, x_label, y_label)

    elif isinstance(fname, list):
      from .graphics.plot_functions import plot_shc_tensor
      es = None
      data = []
      t_ele = []
      labels = []
      if not isinstance(cols, list):
        cols = [cols] * len(fname)
      for fn in fname:
        tag = fn.split('.')[-2][-4:]
        es,shc = read_dos_PAO(fn)
        data.append(shc)
        labels.append(tag)
      plot_shc_tensor(es, data, title, x_lim, y_lim, x_label, y_label, cols, labels, legend)




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import os
import re
import sys
import traceback
import numpy as np
import xml.etree.cElementTree as ET

def read_attribute ( aroot, default_value, attr, atype, alen=1 ):
    txt = aroot.findall(attr)
    read_value = None
    try:
        if len(txt) == alen:
            if atype == 'logical':
                read_value = (str(txt[0].text) == 'T')
            elif atype == 'integer':
                read_value = int(txt[0].text)
            elif atype == 'decimal':
                read_value = float(txt[0].text)
            elif atype == 'array':
                ntxt = aroot.findall(attr+'/a')
                m = len(ntxt)
                if m > 0:
                    n = len(ntxt[0].text.split())
                    read_value = np.zeros((m,n), dtype=float)
                    for i in range(m):
                        line = ntxt[i].text.split()
                        for j in range(n):
                            read_value[i,j] = line[j]
            elif atype == 'string_array':
                tmp_dict = {}
                ntxt = aroot.findall(attr+'/a')
                m = len(ntxt)
                if m > 0:
                  for i in range(m):
                    line = ntxt[i].text.split()
                    tmp_dict[line[0]] = np.asarray(line[1:], dtype=float)
                read_value = tmp_dict

        if atype == 'string':
          if len(txt) == 1:
            read_value = str(txt[0].text)
          elif len(txt) > 0:
            ov = []
            read_value = [ov.append(str(t.text)) for t in txt]
    except: pass
    if read_value is not None:
        return read_value
    else:
        return default_value

def read_inputfile_xml ( fpath, inputfile, data_controller ):

    fname = inputfile
    inputfile = os.path.join(fpath,fname)
    data_arrays = data_controller.data_arrays
    data_attributes = data_controller.data_attributes

    # Control
    fpath = None
    restart = verbose = acbn0 = write2file = write_binary = writedata = use_cuda = writez2pack = False
    shift_type = 1
    shift = 'auto'
    pthr = 0.95
    npool = 1

    # Return Request Definitions
    out_vals = ''

    # Calculation

    # Compare PAO bands with original DFT bands on the original MP mesh
    do_comparison = False

    # Dimensions of the atomic basis for each atom (order must be the same as in the output of projwfc.x)
    naw = np.array([[0,0]])

    # Shell order and degeneracy for SO (order must be the same as in the output of projwfc.x)
    sh = np.array([[]])    # order of shells with l angular momentum
    nl = np.array([[]])    # multiplicity of each l shell

    # External fields
    Efield = np.array([[0,0,0]]) # static electric field (eV)
    Bfield = np.array([[0,0,0]]) # static magnetic firld placeholder: magnetic supercell not implemented!
    HubbardU = np.zeros(32,dtype=float) # non scf ACBN0 calculation
    HubbardU[1:4] = 0.0
    bval = 0 # top valence band number (nelec/2) to correctly shift eigenvalues

    # Bands interpolation along a path from a 1D string of k points
    onedim = False

    # Bands interpolation on a path from the original MP mesh 
    do_bands = False
    ibrav = 0
    dkres = 0.1
    nk    = 2000
    band_path = None
    high_sym_points = np.array([[]])

    # Band topology analysis
    band_topology = False
    eff_mass = False
    spol = 0  # spin
    ipol = 0
    jpol = 0

    # Construct PAO spin-orbit Hamiltonian
    do_spin_orbit = False
    theta = 0.0
    phi = 0.0
    lambda_p = np.array([[0.0]])
    lambda_d = np.array([[0.0]])

    # Hamiltonian interpolation on finer MP mesh
    double_grid = False
    nfft1 = 0
    nfft2 = 0
    nfft3 = 0

    # DOS(PDOS) calculation
    do_dos = False
    do_pdos = False
    emin = -10.
    emax = 2
    delta = 0.01

    # Adaptive smearing
    smearing = 'gauss' # other available values are None or 'm-p'

    # Plot Fermi Surface (spin texture)
    fermisurf = False
    fermi_up = 0.1
    fermi_dw = -0.1
    spintexture = False

    # Tensor components
    # Dielectric function
    d_tensor = np.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2],[2,0],[2,1],[2,2]])
    # Boltzmann transport
    t_tensor = np.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2],[2,0],[2,1],[2,2]])
    # Berry curvature
    a_tensor = np.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2],[2,0],[2,1],[2,2]])
    # Spin Berry curvature
    s_tensor = np.array([[0,0,0],[0,1,0],[0,2,0],[1,0,0],[1,1,0],[1,2,0],[2,0,0],[2,1,0],[2,2,0], \
                  [0,0,1],[0,1,1],[0,2,1],[1,0,1],[1,1,1],[1,2,1],[2,0,1],[2,1,1],[2,2,1], \
                  [0,0,2],[0,1,2],[0,2,2],[1,0,2],[1,1,2],[1,2,2],[2,0,2],[2,1,2],[2,2,2]])

    # Set temperature in eV
    temp = 0.025852  # room temperature

    tmin = 300        # initial temperature in Kelvin
    tmax = 300      # highest temperature in Kelvin
    tstep= 1        # temperature step in Kelvin

    # Boltzmann transport calculation
    Boltzmann = False
    eminBT = -1.
    emaxBT = 1.

    # Evaluate the Carrier Concentration
    carrier_conc = False

    # Dielectric function calculation
    epsilon = False
    metal = False
    kramerskronig = True
    epsmin=0.0
    epsmax=10.0
    ne = 500

    # Critical points
    critical_points = False

    # Berry curvature and AHC
    Berry = False
    eminAH = -1.0
    emaxAH = 1.0
    ac_cond_Berry = False

    # Spin Berry curvature and SHC
    spin_Hall = False
    eminSH = -1.0
    emaxSH = 1.0
    ac_cond_spin = False

    # symmetrization and wedge -> grid
    expand_wedge  = True
    symmetrize    = False
    symm_thresh   = 1.e-6
    symm_max_iter = 16

    tree = ET.parse(inputfile)
    aroot = tree.getroot()

    # Read String Input Values
    data_attributes['fpath'] = read_attribute(aroot, fpath, 'fpath', 'string')
    data_attributes['savedir'] = data_attributes['fpath']
    data_attributes['shift'] = read_attribute(aroot, shift, 'shift', 'string')
    try:
        data_attributes['shift'] = read_attribute(aroot, shift, 'shift', 'decimal')
    except:
        pass
    data_attributes['out_vals'] = read_attribute(aroot, out_vals, 'out_vals', 'string')
    smearing = read_attribute(aroot, smearing, 'smearing', 'string')
    if smearing == 'None':
        smearing = None
    data_attributes['smearing'] = smearing

    # Read Logical Input Values
    data_attributes['restart'] = read_attribute(aroot, restart, 'restart', 'logical')
    data_attributes['writez2pack'] = read_attribute(aroot, writez2pack, 'writez2pack', 'logical')
    data_attributes['verbose'] = read_attribute(aroot, verbose, 'verbose', 'logical')
    data_attributes['acbn0'] = read_attribute(aroot, acbn0, 'non_ortho', 'logical')
    data_attributes['write2file'] = read_attribute(aroot, write2file, 'write2file', 'logical')
    data_attributes['write_binary'] = read_attribute(aroot, write_binary, 'write_binary', 'logical')
    data_attributes['writedata'] = read_attribute(aroot, writedata, 'writedata', 'logical')
    data_attributes['use_cuda'] = read_attribute(aroot, use_cuda, 'use_cuda', 'logical')
    data_attributes['do_comparison'] = read_attribute(aroot, do_comparison, 'do_comparison', 'logical')
    data_attributes['onedim'] = read_attribute(aroot, onedim, 'onedim', 'logical')
    data_attributes['do_bands'] = read_attribute(aroot, do_bands, 'do_bands', 'logical')
    data_attributes['band_topology'] = read_attribute(aroot, band_topology, 'band_topology', 'logical')
    data_attributes['eff_mass'] = read_attribute(aroot, eff_mass, 'eff_mass_topology', 'logical')
    data_attributes['do_spin_orbit'] = read_attribute(aroot, do_spin_orbit, 'do_spin_orbit', 'logical')
    data_attributes['double_grid'] = read_attribute(aroot, double_grid, 'double_grid', 'logical')
    data_attributes['do_dos'] = read_attribute(aroot, do_dos, 'do_dos', 'logical')
    data_attributes['do_pdos'] = read_attribute(aroot, do_pdos, 'do_pdos', 'logical')
    data_attributes['fermisurf'] = read_attribute(aroot, fermisurf, 'fermisurf', 'logical')
    data_attributes['spintexture'] = read_attribute(aroot, spintexture, 'spintexture', 'logical')
    data_attributes['Boltzmann'] = read_attribute(aroot, Boltzmann, 'Boltzmann', 'logical')
    data_attributes['carrier_conc'] = read_attribute(aroot, carrier_conc, 'carrier_conc', 'logical')
    data_attributes['epsilon'] = read_attribute(aroot, epsilon, 'epsilon', 'logical')
    data_attributes['metal'] = read_attribute(aroot, metal, 'metal', 'logical')
    data_attributes['kramerskronig'] = read_attribute(aroot, kramerskronig, 'kramerskronig', 'logical')
    data_attributes['critical_points'] = read_attribute(aroot, critical_points, 'critical_points', 'logical')
    data_attributes['Berry'] = read_attribute(aroot, Berry, 'Berry', 'logical')
    data_attributes['ac_cond_Berry'] = read_attribute(aroot, ac_cond_Berry, 'ac_cond_Berry', 'logical')
    data_attributes['spin_Hall'] = read_attribute(aroot, spin_Hall, 'spin_Hall', 'logical')
    data_attributes['ac_cond_spin'] = read_attribute(aroot, ac_cond_spin, 'ac_cond_spin', 'logical')


    # Read Integer Input Values
    data_attributes['shift_type'] = read_attribute(aroot, shift_type, 'shift_type', 'integer')
    data_attributes['npool'] = read_attribute(aroot, npool, 'npool', 'integer')
    data_attributes['bval'] = read_attribute(aroot, bval, 'bval', 'integer')
    data_attributes['ibrav'] = read_attribute(aroot, ibrav, 'ibrav', 'integer')
    data_attributes['band_path'] = read_attribute(aroot, band_path, 'band_path', 'string')
    data_attributes['ne'] = read_attribute(aroot, ne, 'ne', 'integer')
    data_attributes['nk'] = read_attribute(aroot, nk, 'nk', 'integer')
    data_attributes['spol'] = read_attribute(aroot, spol, 'spol', 'integer')
    data_attributes['ipol'] = read_attribute(aroot, ipol, 'ipol', 'integer')
    data_attributes['jpol'] = read_attribute(aroot, jpol, 'jpol', 'integer')
    data_attributes['nfft1'] = read_attribute(aroot, nfft1, 'nfft1', 'integer')
    data_attributes['nfft2'] = read_attribute(aroot, nfft2, 'nfft2', 'integer')
    data_attributes['nfft3'] = read_attribute(aroot, nfft3, 'nfft3', 'integer')

    # Read Decimal Input Values
    data_attributes['pthr'] = read_attribute(aroot, pthr, 'pthr', 'decimal')
    data_attributes['dkres'] = read_attribute(aroot, dkres, 'dkres', 'decimal')
    data_attributes['theta'] = read_attribute(aroot, theta, 'theta', 'decimal')
    data_attributes['phi'] = read_attribute(aroot, phi, 'phi', 'decimal')
    data_attributes['emin'] = read_attribute(aroot, emin, 'emin', 'decimal')
    data_attributes['emax'] = read_attribute(aroot, emax, 'emax', 'decimal')
    data_attributes['delta'] = read_attribute(aroot, delta, 'delta', 'decimal')
    data_attributes['fermi_up'] = read_attribute(aroot, fermi_up, 'fermi_up', 'decimal')
    data_attributes['fermi_dw'] = read_attribute(aroot, fermi_dw, 'fermi_dw', 'decimal')
    data_attributes['temp'] = read_attribute(aroot, temp, 'temp', 'decimal')
    data_attributes['tmin'] = read_attribute(aroot, tmin, 'tmin', 'decimal')
    data_attributes['tmax'] = read_attribute(aroot, tmax, 'tmax', 'decimal')
    data_attributes['tstep'] = read_attribute(aroot, tstep, 'tstep', 'decimal')
    data_attributes['epsmin'] = read_attribute(aroot, epsmin, 'epsmin', 'decimal')
    data_attributes['epsmax'] = read_attribute(aroot, epsmax, 'epsmax', 'decimal')
    data_attributes['eminAH'] = read_attribute(aroot, eminAH, 'eminAH', 'decimal')
    data_attributes['emaxAH'] = read_attribute(aroot, emaxAH, 'emaxAH', 'decimal')
    data_attributes['eminSH'] = read_attribute(aroot, eminSH, 'eminSH', 'decimal')
    data_attributes['emaxSH'] = read_attribute(aroot, emaxSH, 'emaxSH', 'decimal')
    data_attributes['eminBT'] = read_attribute(aroot, eminBT, 'eminBT', 'decimal')
    data_attributes['emaxBT'] = read_attribute(aroot, emaxBT, 'emaxBT', 'decimal')

    # Read Array Input Values
    data_arrays['naw'] = read_attribute(aroot, naw, 'naw', 'array')[0].astype(int)
    data_arrays['sh'] = [int(i) for i in read_attribute(aroot, sh, 'sh', 'array')[0]]
    data_arrays['nl'] = [int(i) for i in read_attribute(aroot, nl, 'nl', 'array')[0]]
    if len(data_arrays['sh']) == 0 or len(data_arrays['sh']) != len(data_arrays['nl']):
      del data_arrays['sh']
      del data_arrays['nl']
    data_arrays['Efield'] = read_attribute(aroot, Efield, 'Efield', 'array')[0]
    data_arrays['Bfield'] = read_attribute(aroot, Bfield, 'Bfield', 'array')[0]
    data_arrays['HubbardU'] = read_attribute(aroot, HubbardU, 'HubbardU', 'array')[0]
    data_arrays['lambda_p'] = read_attribute(aroot, lambda_p, 'lambda_p', 'array')[0]
    data_arrays['lambda_d'] = read_attribute(aroot, lambda_d, 'lambda_d', 'array')[0]
    data_arrays['d_tensor'] = read_attribute(aroot, d_tensor, 'd_tensor', 'array').astype(int)
    data_arrays['t_tensor'] = read_attribute(aroot, t_tensor, 't_tensor', 'array').astype(int)
    data_arrays['a_tensor'] = read_attribute(aroot, a_tensor, 'a_tensor', 'array').astype(int)
    data_arrays['s_tensor'] = read_attribute(aroot, s_tensor, 's_tensor', 'array').astype(int)
    data_arrays['high_sym_points'] = read_attribute(aroot, high_sym_points, 'high_sym_points', 'string_array')

    # symmetrization and wedge -> grid
    data_attributes['expand_wedge'] = read_attribute(aroot, expand_wedge, 'expand_wedge', 'logical')
    data_attributes['symmetrize'] = read_attribute(aroot, symmetrize, 'symmetrize', 'logical')
    data_attributes['symm_thresh'] = read_attribute(aroot,symm_thresh, 'symm_thresh', 'decimal')
    data_attributes['symm_max_iter'] = read_attribute(aroot, symm_max_iter, 'symm_max_iter', 'integer')




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

from .smearing import intgaussian,gaussian
from .smearing import intmetpax,metpax
from .do_atwfc_proj import *
from .do_dipole import *

def do_dielectric_tensor ( data_controller, ene, from_wfc):
  from .constants import LL, RYTOEV

  arrays,attributes = data_controller.data_dicts()
  d_tensor = arrays['d_tensor']
  esize = ene.size

  nspin = attributes['nspin']

  if from_wfc == None:
    pass
  elif from_wfc == 'external':
    nbnds = attributes['nbnds']
    nkpnts = attributes['nkpnts']
    nspin = attributes['nspin']
    arrays['pksp'] = np.empty((nkpnts,3,nbnds,nbnds,nspin),dtype=np.complex128)
    for ispin in range(nspin):
      for ik in range(nkpnts):
        arrays['pksp'][ik,:,:,:,ispin] = calc_dipole(arrays,attributes, ik, ispin, arrays['b_vectors'])
  elif from_wfc == 'internal':
    nbnds = attributes['nawf']
    nkpnts = attributes['nkpnts']
    nspin = attributes['nspin']
    arrays['pksp'] = np.empty((nkpnts,3,nbnds,nbnds,nspin),dtype=np.complex128)
    for ispin in range(nspin):
      for ik in range(nkpnts):
        arrays['pksp'][ik,:,:,:,ispin] = calc_dipole_internal(data_controller, ik, ispin)
        # for n in range(attributes['nawf']):
        #   for m in range(attributes['nawf']):
        #     for l in range(3):
        #       arrays['pksp'][ik,l,n,m,ispin] = 1j*arrays['pksp'][ik,l,n,m,ispin]*(arrays['E_k'][ik,n]-arrays['E_k'][ik,m])
  else:
    raise Exception('ERROR: no dipole mode specified')

  if nspin == 1:
    for n in range(d_tensor.shape[0]):
      ipol = d_tensor[n][0]
      jpol = d_tensor[n][1]

      epsi,epsr,eels,ieps = do_epsilon(data_controller, ene, 0, ipol, jpol, from_wfc)
      # Write files
      indices = (LL[ipol], LL[jpol])
      for ep,es in [(epsi,'epsi'),(epsr,'epsr'),(eels,'eels'),(ieps,'ieps')]:
        fn = '%s_%s%s.dat'%((es,)+indices)
        data_controller.write_file_row_col(fn, ene, ep)

      if rank == 0 and ipol == jpol:
        renorm = np.sqrt((2./np.pi)*np.trapezoid(epsi*ene,x=ene))
        component = LL[ipol]+LL[jpol]
        print('Component', component, ', plasmon frequency = ',renorm,'eV')

  else:
    for n in range(d_tensor.shape[0]):
      ipol = d_tensor[n][0]
      jpol = d_tensor[n][1]

      epsi_0,epsr_0,eels_0,ieps_0 = do_epsilon(data_controller, ene, 0, ipol, jpol, from_wfc)
      epsi_1,epsr_1,eels_1,ieps_1 = do_epsilon(data_controller, ene, 1, ipol, jpol, from_wfc)
      # Write files
      indices = (LL[ipol], LL[jpol],0)
      for ep,es in [(epsi_0,'epsi'),(epsr_0,'epsr'),
                         (eels_0,'eels'),(ieps_0,'ieps')]:
        fn = '%s_%s%s_%d.dat'%((es,)+indices)
        data_controller.write_file_row_col(fn, ene, ep)
      indices = (LL[ipol], LL[jpol],1)
      for ep,es in [(epsi_1,'epsi'),(epsr_1,'epsr'),(eels_1,'eels'),(ieps_1,'ieps')]:
        fn = '%s_%s%s_%d.dat'%((es,)+indices)
        data_controller.write_file_row_col(fn, ene, ep)

      if rank == 0 and ipol == jpol:
        epsi = epsi_0 + epsi_1
        renorm = np.sqrt((2./np.pi)*np.trapezoid(epsi*ene,x=ene))
        component = LL[ipol]+LL[jpol]
        print('Component', component, ', plasmon frequency = ',renorm,'eV')
        

def do_jdos( data_controller, ene, jdos_smeartype):
  _,attributes = data_controller.data_dicts()
  esize = ene.size

  nspin = attributes['nspin']
  if nspin == 1:
    jdos_aux = jdos_loop(data_controller, ene, 0, jdos_smeartype)
    jdos = np.zeros(esize, dtype=float)
    comm.Allreduce(jdos_aux, jdos, op=MPI.SUM)
    jdos_aux = None

    fn = 'jdos.dat'
    data_controller.write_file_row_col(fn, ene, jdos)

    if rank == 0:
      print('Integration over JDOS = ', (np.trapezoid(jdos,x=ene)))
  else:
    jdos_aux0 = jdos_loop(data_controller, ene, 0, jdos_smeartype)
    jdos_aux1 = jdos_loop(data_controller, ene, 1, jdos_smeartype)
    jdos0 = np.zeros(esize, dtype=float)
    jdos1 = np.zeros(esize, dtype=float)
    comm.Allreduce(jdos_aux0, jdos0, op=MPI.SUM)
    comm.Allreduce(jdos_aux1, jdos1, op=MPI.SUM)
    jdos_aux0 = None
    jdos_aux1 = None

    fn0 = 'jdos_0.dat'
    data_controller.write_file_row_col(fn0, ene, jdos0)
    fn1 = 'jdos_1.dat'
    data_controller.write_file_row_col(fn1, ene, jdos1)

    if rank == 0:
      print('Integration over JDOS = ', (np.trapezoid(jdos0+jdos1,x=ene)))


def do_epsilon ( data_controller, ene, ispin, ipol, jpol, from_wfc):
  from .constants import BOHR_RADIUS_ANGS, RYTOEV,ELECTRONVOLT_SI

  # Compute the dielectric tensor

  _,attributes = data_controller.data_dicts()

  esize = ene.size
  if ene[0] == 0.:
    ene[0] = .00001
  
  if from_wfc == None:
    # 8.8541878188e-12 = \epsilon_0
    factor = ELECTRONVOLT_SI*(1e10)/(8.8541878188e-12)*BOHR_RADIUS_ANGS**2\
      /attributes['nkpnts']/(attributes['omega']*BOHR_RADIUS_ANGS**3)
  elif from_wfc == 'external':
    factor = 2*(np.pi/attributes['alat'])**2*RYTOEV**3\
      *64.0*np.pi/(attributes['omega']*attributes['nkpnts'])
  elif from_wfc == 'internal':
    # factor = ELECTRONVOLT_SI*(1e10)/(8.8541878188e-12)*BOHR_RADIUS_ANGS**2\
    #   /attributes['nkpnts']/(attributes['omega']*BOHR_RADIUS_ANGS**3)
    factor = 2*(np.pi/attributes['alat'])**2*RYTOEV**3\
      *64.0*np.pi/(attributes['omega']*attributes['nkpnts'])
  else:
    raise Exception('ERROR: no dipole mode specified')
     
  #=======================
  # EPS
  #=======================

  epsi_aux,epsr_aux = eps_loop(data_controller, ene, ispin, ipol, jpol, from_wfc)
  epsi = np.zeros(esize, dtype=float)
  comm.Allreduce(epsi_aux, epsi, op=MPI.SUM)
  epsi_aux = None

  epsr = np.zeros(esize, dtype=float)
  comm.Allreduce(epsr_aux, epsr, op=MPI.SUM)
  epsr_aux = None


  ### TNeeds revision. Each processor is allocating zeros here, when only rank 0 needs it. 
  ### Can be condensed
  
  ieps = np.zeros(esize, dtype=float)
    
  epsi *= factor
  epsr =  1.*(ipol==jpol) + epsr*factor 
  eels = epsi/(epsi**2+epsr**2)
  for i in range(esize):
    for j in range(1,esize):
        ieps[i] += ene[j]*epsi[j]/(ene[i]**2+ene[j]**2)
  ieps = 1.0 + (2./np.pi)*ieps*(ene[3]-ene[2])

  return(epsi, epsr, eels, ieps)


def eps_loop ( data_controller, ene, ispin, ipol, jpol, from_wfc):

  orig_over_err = np.geterr()['over']
  np.seterr(over='raise')

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  if from_wfc == None or from_wfc == 'internal':
    bndmax = attributes['bnd']
    Ek = arrays['E_k'][:,:bndmax,ispin]
  elif from_wfc == 'external':
    bndmax = attributes['bnd']
    Ek = np.swapaxes(arrays['my_eigsmat'][:,:,ispin],0,1)
  else:
    raise Exception('ERROR: no dipole mode specified')
  # 

  intersmear = attributes['delta']  
  smearing = attributes['smearing']
  degauss = attributes['degauss']
  
  # spin_factor = 2 if attributes['nspin']==1 and not attributes['dftSO'] else 1
  spin_factor = 2 if attributes['nspin']==1 else 1
  Ef = 1.e-9

  epsi = np.zeros(esize, dtype=float)
  epsr = np.zeros(esize, dtype=float)
    
  if smearing == None or attributes['insulator']:
      if rank == 0: print("No smearing, fixed occupation")
  else:
    if 'deltakp' in arrays:  # check whether adaptive smearing is used
      degauss = arrays['deltakp'][:,:bndmax,ispin] 
      if rank == 0: print("Using adaptive smearing")
    else:
      degauss = attributes['degauss']
      if rank == 0: print("Using fixed smearing = %.3f eV" %degauss)

  if smearing == None or attributes['insulator']:
    fn = spin_factor*(Ek <= Ef)  # fixed occupation for insulator, no smearing
  elif smearing == 'gauss':
    
    fn = spin_factor*intgaussian(Ek, Ef, degauss) 
  else: # smearing == 'm-p':

    fn = spin_factor*intmetpax(Ek, Ef, degauss)

  th0 = 1.e-3*spin_factor
  th1 = 0.5e-4*spin_factor
  if attributes['dftSO'] and from_wfc == None: 
    fac = 1
  else:
    fac = 2
  for ik in range(fn.shape[0]):
    for iband2 in range(bndmax):
       for iband1 in range(bndmax):
          if iband1 != iband2:
             E_diff_nm = Ek[ik,iband2] - Ek[ik,iband1]
             f_nm =  fn[ik,iband2]-fn[ik,iband1]
             if np.abs(f_nm) > th0 and fn[ik,iband1] > th1 and fn[ik,iband2] < spin_factor:
                pksp2 = np.real(arrays['pksp'][ik,ipol,iband1,iband2,ispin]\
                        *arrays['pksp'][ik,jpol,iband2,iband1,ispin])
                # pksp2 in unit of (AU*eV)^2
                epsi[:] += fac*pksp2*intersmear*ene[:]*fn[ik,iband1]\
                /(((E_diff_nm**2-ene[:]**2)**2+intersmear**2*ene[:]**2)*(E_diff_nm))
                epsr[:] += fac*pksp2*(E_diff_nm**2-ene[:]**2)*fn[ik,iband1]\
                /(((E_diff_nm**2-ene[:]**2)**2+intersmear**2*ene[:]**2)*(E_diff_nm))

  
  if not attributes['insulator']:
    epsi_metal = np.zeros_like(epsi)
    epsr_metal = np.zeros_like(epsr)

    if smearing == 'gauss':
      fnF = spin_factor*gaussian(Ek, Ef, degauss)
    elif smearing == 'm-p':
      fnF = spin_factor*metpax(Ek, Ef, degauss)
    else:
      print("Smearing is None for a metal, switching to gaussian smearing")
      fnF = spin_factor*gaussian(Ek, Ef, degauss)

    intrasmear = attributes['intrasmear']

    for ik in range(fn.shape[0]):
      for iband1 in range(bndmax):
          pksp2 = np.real(arrays['pksp'][ik,ipol,iband1,iband1,ispin]\
                          *arrays['pksp'][ik,jpol,iband1,iband1,ispin])
          epsi_metal[:] += pksp2*intrasmear*ene[:]*fnF[ik,iband1]/(ene[:]**4+intrasmear**2*ene[:]**2)
          epsr_metal[:] -= pksp2*fnF[ik,iband1]*ene[:]**2/(ene[:]**4+intrasmear**2*ene[:]**2)

    if from_wfc != None:
      from .constants import RYTOEV
      epsi_metal /= RYTOEV/4
      epsr_metal /= RYTOEV/4

    epsi += epsi_metal
    epsr += epsr_metal

  np.seterr(over=orig_over_err)
  return(epsi, epsr)


def jdos_loop(data_controller, ene, ispin, jdos_smeartype):

  arrays,attributes = data_controller.data_dicts()
  intersmear = attributes['delta']  
  smearing = attributes['smearing']
  esize = ene.size
  bndmax = attributes['bnd']
  Ek = arrays['E_k'][:,:bndmax,ispin]
  # bndmax = attributes['nbnds']
  # Ek = np.swapaxes(arrays['my_eigsmat'][:,:,ispin],0,1)
  nkpnts = Ek.shape[0]
  jdos = np.zeros(esize, dtype=float)
  Ef = 1.e-9
  degauss = attributes['degauss']

  # spin_factor = 2 if attributes['nspin']==1 and not attributes['dftSO'] else 1
  spin_factor = 2 if attributes['nspin']==1 else 1
  if smearing == None or attributes['insulator']:
    fn = spin_factor*(Ek <= Ef)  # fixed occupation for insulator, no smearing
  elif smearing == 'gauss':
    fn = spin_factor*intgaussian(Ek, Ef, degauss) 
  else: # smearing == 'm-p':
    fn = spin_factor*intmetpax(Ek, Ef, degauss)

  # count = 0.0
  if jdos_smeartype == 'gauss':

    for ik in range(nkpnts):
      for iband2 in range(bndmax):
        for iband1 in range(bndmax):
            E_diff_nm = Ek[ik,iband2] - Ek[ik,iband1]
            if fn[ik,iband1] > 1.e-4 and fn[ik,iband2] < 2.0 and E_diff_nm > 1e-10:
              f_nm = fn[ik,iband1]-fn[ik,iband2]
              jdos += f_nm * gaussian(E_diff_nm, ene, intersmear) 
              # count += f_nm

  elif jdos_smeartype == 'lorentz':
  
    for ik in range(nkpnts):
      for iband2 in range(bndmax):
        for iband1 in range(bndmax):
            E_diff_nm = Ek[ik,iband2] - Ek[ik,iband1]
            if fn[ik,iband1] > 1.e-4 and fn[ik,iband2] < 2.0 and E_diff_nm > 1e-10:
              f_nm = fn[ik,iband1]-fn[ik,iband2]
              jdos +=  f_nm * intersmear / (np.pi*((E_diff_nm-ene)**2+intersmear**2))
              # count += f_nm

  else:
    raise ValueError("jdos_smeartype must be 'gauss' or 'lorentz' ")
  jdos /= nkpnts

  return jdos

   

"""
def epsr_kramerskronig ( data_controller, ene, epsi ):
  from .smearing import intmetpax
  from scipy.integrate import simpson
  from .communication import load_balancing

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  de = ene[1] - ene[0]

  epsr = np.zeros(esize, dtype=float)

  ini_ie,end_ie = load_balancing(comm.Get_size(), rank, esize)

  # Range checks for Simpson Integrals
  if end_ie == ini_ie:
    return
  if ini_ie < 3:
    ini_ie = 3
  if end_ie == esize:
    end_ie = esize-1

  f_ene = intmetpax(ene, attributes['shift'], 1.)
  for ie in range(ini_ie, end_ie):
    I1 = simpson(ene[1:(ie-1)]*de*epsi[1:(ie-1)]*f_ene[1:(ie-1)]/(ene[1:(ie-1)]**2-ene[ie]**2))
    I2 = simpson(ene[(ie+1):esize]*de*epsi[(ie+1):esize]*f_ene[(ie+1):esize]/(ene[(ie+1):esize]**2-ene[ie]**2))
    epsr[ie] = 2.*(I1+I2)/np.pi

  return epsr
"""


#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

# This module is taken from AFLOWpi
import numpy as np
import sys,os
import copy


from mpi4py import MPI
from mpi4py.MPI import ANY_SOURCE

# initialize parallel execution
comm=MPI.COMM_WORLD
rank = comm.Get_rank()



def _getHighSymPoints ( ibrav, alat, cellOld ):
  '''
  Searching for the ibrav number in the input file for the calculation
  to determine the path for the band structure calculation

  Arguments:
      oneCalc (dict): a dictionary containing properties about the AFLOWpi calculation

  Keyword Arguments:
      ID (str): ID string for the particular calculation and step

  Returns:
      special_points (list): list of the HSP names
      band_path (str): path in string form

  '''


###############################################################################
###############################################################################

  #get a,b,c of QE convention conventional cell from primitive lattice vecs
  if ibrav == 1:
    a=np.abs(cellOld[0][0])*2.0
  if ibrav == 2:
    a=np.abs(cellOld[0][0])*2.0
  if ibrav == 3:
    a=np.abs(cellOld[0][0])*2.0
  if ibrav == 4:
    a=np.abs(cellOld[0][0])
    b=a
    c=np.abs(cellOld[2][2]*a)
  if ibrav == 5:
    a=np.sqrt(cellOld[0].dot(cellOld[0].T))
    alpha=np.arccos(cellOld[0].dot(cellOld[1].T)/(a**2))
    alpha_deg = alpha*180.0/np.pi
  if ibrav == 6:
    a=np.abs(cellOld[0][0])
    c=np.abs(cellOld[2][2]*a)
  if ibrav == 7:
    a=np.abs(cellOld[0][0])*2.0
    c=np.abs(cellOld[2][2]*a)*2.0
  if ibrav == 9:
    a=np.abs(cellOld[0][0])*2.0
    b=np.abs(cellOld[1][1])*2.0*a
    c=np.abs(cellOld[2][2])*a
  if ibrav == 10:
    a=np.abs(cellOld[0][0])*2.0
    b=np.abs(cellOld[1][1])*2.0*a
    c=np.abs(cellOld[2][2])*2.0*a
  if ibrav == 11:
    a=np.abs(cellOld[0][0])*2.0
    b=np.abs(cellOld[1][1])*2.0*a
    c=np.abs(cellOld[2][2])*2.0*a
  if ibrav == 12:
    '''swapped compared to AFLOW..ABC->CAB'''
    a=np.sqrt(cellOld[0].dot(cellOld[0].T))
    b=np.sqrt(cellOld[1].dot(cellOld[1].T))
    c=np.sqrt(cellOld[2].dot(cellOld[2].T))
    alpha=np.arccos(cellOld[1].dot(cellOld[2].T)/(c*b))
    beta =np.arccos(cellOld[0].dot(cellOld[2].T)/(a*c))
    gamma=np.arccos(cellOld[0].dot(cellOld[1].T)/(a*b))


  if ibrav==14:
    a=np.sqrt(cellOld[0].dot(cellOld[0].T))
    b=np.sqrt(cellOld[1].dot(cellOld[1].T))
    c=np.sqrt(cellOld[2].dot(cellOld[2].T))
    alpha=np.arccos(cellOld[1].dot(cellOld[2].T)/(c*b))
    beta =np.arccos(cellOld[0].dot(cellOld[2].T)/(a*c))
    gamma=np.arccos(cellOld[0].dot(cellOld[1].T)/(a*b))


  if   ibrav==1:  ibrav_var =  'CUB'
  elif ibrav==2:  ibrav_var =  'FCC'
  elif ibrav==3:  ibrav_var =  'BCC'
  elif ibrav==4:  ibrav_var =  'HEX'
  elif ibrav==6:  ibrav_var =  'TET'
  elif ibrav==8:  ibrav_var =  'ORC'
  elif ibrav==9:  ibrav_var =  'ORCC'
  elif ibrav==11: ibrav_var =  'ORCI'
  elif ibrav==12: ibrav_var =  'MCL'

  elif ibrav==5:
    if alpha_deg   < 90.0: ibrav_var = 'RHL1'
    elif alpha_deg > 90.0: ibrav_var = 'RHL2'
  elif ibrav==7:
    if(c < a):   ibrav_var =  'BCT1'
    elif(c > a): ibrav_var =  'BCT2'
    else:    ibrav_var =  'BCC'
  elif ibrav==10:

    if  (1.0/a**2 > 1.0/b**2+1.0/c**2): ibrav_var =  'ORCF1'
    elif  np.isclose(1.0/a**2, 1.0/b**2+1.0/c**2,1.e-2): ibrav_var =  'ORCF3'
    elif  (1.0/a**2 < 1.0/b**2+1.0/c**2): ibrav_var =  'ORCF2'

  elif(int(ibrav)==14):
    minAngle = np.amin([alpha,beta,gamma])
    maxAngle = np.amax([alpha,beta,gamma])
    if alpha==90.0 or beta==90.0 or gamma==90.0:
      if alpha>=90.0 or beta>=90.0 or gamma>=90.0: ibrav_var =  'TRI2A'
      if alpha<=90.0 or beta<=90.0 or gamma<=90.0: ibrav_var =  'TRI2B'
    elif minAngle>90.0:                ibrav_var =  'TRI1A'
    elif maxAngle<90:                ibrav_var =  'TRI1B'
###############################################################################
###############################################################################
  if ibrav_var=='CUB':
    band_path = 'gG-X-M-gG-R-X|M-R'
    special_points = {'gG'   : (0.0, 0.0, 0.0),
               'M'   : (0.5, 0.5, 0.0),
               'R'   : (0.5, 0.5, 0.5),
               'X'   : (0.0, 0.5, 0.0)}
               
  if ibrav_var=='FCC':
    band_path = 'gG-X-W-K-gG-L-U-W-L-K|U-X'
    special_points = {'gG'   : (0.0, 0.0, 0.0),
              'K'  : (0.375, 0.375, 0.750),
              'L'  : (0.5, 0.5, 0.5),
              'U'  : (0.625, 0.250, 0.625),
              'W'  : (0.5, 0.25, 0.75),
              'X'  : (0.5, 0.0, 0.5)}
              
  if ibrav_var=='BCC':
    band_path = 'gG-H-N-gG-P-H|P-N'
    special_points = {'gG'   : (0, 0, 0),
              'H'  : (0.5, -0.5, 0.5),
              'P'  : (0.25, 0.25, 0.25,), 
              'N'  : (0.0, 0.0, 0.5)}
      
  if ibrav_var=='HEX':
    band_path = 'gG-M-K-gG-A-L-H-A|L-M|K-H'
    special_points = {'gG'   : (0, 0, 0),
              'A'  : (0.0, 0.0, 0.5),
              'H'  : (1.0/3.0, 1.0/3.0, 0.5),
              'K'  : (1.0/3.0, 1.0/3.0, 0.0),
              'L'  : (0.5, 0.0, 0.5),
              'M'  : (0.5, 0.0, 0.0)}
    
  if ibrav_var=='RHL1':
    eta = (1.0 + 4.0*np.cos(alpha))/(2.0 + 4.0*np.cos(alpha))
    

    nu =0.75-eta/2.0
    band_path = 'gG-L-B1|B-Z-gG-X|Q-F-P1-Z|L-P'
    special_points = {'gG'   : (0.0, 0.0, 0.0),
              'B'  : (eta, 0.5, 1.0-eta),
              'B1'   : (0.5, 1.0-eta, eta-1.0),
              'F'  : (0.5, 0.5, 0.0),
              'L'  : (0.5, 0.0, 0.0),
              'L1'   : (0.0, 0.0, -0.5),
              'P'  : (eta, nu, nu),
              'P1'   : (1.0-nu, 1.0-nu, 1.0-eta),
              'P2'   : (nu, nu, eta-1.0),
              'Q'  : (1.0-nu, nu, 0.0),
              'X'  : (nu, 0.0, -nu),
              'Z'  : (0.5, 0.5, 0.5)}
     
  if ibrav_var=='RHL2':
    eta=1.0/(2.0*np.tan(alpha/2.0)**2)
    nu =0.75-eta/2.0
    band_path = 'gG-P-Z-Q-gG-F-P1-Q1-L-Z'
    special_points = {'gG'   : (0.0, 0.0, 0.0),
              'F'  : (0.5, -0.5, 0.0),
              'L'  : (0.5, 0.0, 0.0),
              'P'  : (1.0-nu, -nu, 1.0-nu),
              'P1'   : (nu, nu-1.0, nu-1.0),
              'Q'  : (eta, eta, eta),
              'Q1'   : (1.0-eta, -eta, -eta),
              'Z'  : (0.5, -0.5, 0.5)} 

  if ibrav_var=='TET':
    band_path = 'gG-X-M-gG-Z-R-A-Z|X-R|M-A'
    special_points = {'gG'   : (0.0, 0.0, 0.0),
              'A'  : (0.5, 0.5, 0.5),
              'M'  : (0.5, 0.5, 0.0),
              'R'  : (0.0, 0.5, 0.5),
              'X'  : (0.0, 0.5, 0.0),
              'Z'  : (0.0, 0.0, 0.5)}

  if ibrav_var=='BCT1':
     eta = (1.0+(c/a)**2)/4.0
     band_path = 'gG-X-M-gG-Z-P-N-Z1-M|X-P'
     special_points = {'gG'  : (0.0, 0.0, 0.0),
             'M'   : (-0.5, 0.5, 0.5),
             'N'   : (0.0, 0.5, 0.0),
             'P'   : (0.25, 0.25, 0.25),
             'X'   : (0.0, 0.0, 0.5),
             'Z'   : (eta, eta, -eta),
             'Z1'  : (-eta, 1.0-eta, eta)}
     
  if ibrav_var=='BCT2':
     band_path = 'gG-X-Y-gS-gG-Z-gS1-N-P-Y1-Z|X-P'
     eta = (1.0+(a/c)**2)/4.0
     zeta = 0.5*(a/c)**2

     special_points = {'gG'  : (0.0, 0.0, 0.0),
             'N'   : (0.0, 0.5, 0.0),
             'P'   : (0.25, 0.25, 0.25),
             'gS'  : (-eta, eta, eta),
             'gS1'   : (eta, 1-eta, -eta),
             'X'   : (0.0, 0.0, 0.5),
             'Y'   : (-zeta, zeta, 0.5),
             'Y1'  : (0.5, 0.5, -zeta),
             'Z'   : (0.5, 0.5, -0.5)}
     
  if ibrav_var=='ORC':
     band_path = 'gG-X-S-Y-gG-Z-U-R-T-Z|Y-T|U-X|S-R'
     special_points = {'gG'  : (0.0, 0.0, 0.0),
               'R'   : (0.5, 0.5, 0.5),
               'S'   : (0.5, 0.5, 0.0),
               'T'   : (0.0, 0.5, 0.5),
               'U'   : (0.5, 0.0, 0.5),
               'X'   : (0.5, 0.0, 0.0),
               'Y'   : (0.0, 0.5, 0.0),
               'Z'   : (0.0, 0.0, 0.5)}

  if ibrav_var=='ORCF1':
     band_path = 'gG-Y-T-Z-gG-X-A1-Y|T-X1|X-A-Z|L-gG'
     sorted_lat=np.sort([a,b,c])
     a=sorted_lat[0]
     b=sorted_lat[1]
     c=sorted_lat[2]
     eta =(1.0+(a/b)**2+(a/c)**2)/4.0
     zeta=(1.0+(a/b)**2-(a/c)**2)/4.0
     special_points = {'gG'  : (0.0, 0.0, 0.0),
             'A'   : (0.5, 0.5 + zeta, zeta),
             'A1'  : (0.5, 0.5-zeta, 1.0-zeta),
             'L'   : (0.5, 0.5, 0.5),
             'T'   : (1.0, 0.5, 0.5),
             'X'   : (0.0, eta, eta),
             'X1'  : (1.0, 1.0-eta, 1.0-eta),
             'Y'   : (0.5, 0.0, 0.5),
             'Z'   : (0.5, 0.5, 0.0)}

  if ibrav_var=='ORCF2':
     band_path = 'gG-Y-C-D-X-gG-Z-D1-H-C|C1-Z|X-H1|H-Y|L-gG'
     # sorted_lat=np.sort([a,b,c])
     # a=sorted_lat[0]
     # b=sorted_lat[1]
     # c=sorted_lat[2]
     eta =(1.0+(a/b)**2-(a/c)**2)/4.0
     phi =(1.0+(c/b)**2-(c/a)**2)/4.0
     delta =(1.0+(b/a)**2-(b/c)**2)/4.0

     special_points = {'gG'  : (0.0, 0.0, 0.0),
             'C'   : (0.5, 0.5-eta, 1.0-eta),
             'C1'  : (0.5, 0.5+eta, eta),
             'D'   : (0.5-delta, 0.5, 1.0-delta),
             'D1'  : (0.5+delta, 0.5, delta),
             'L'   : (0.5, 0.5, 0.5),
             'H'   : (1.0-phi, 0.5-phi, 0.5),
             'H1'  : (phi, 0.5+phi, 0.5),
             'X'   : (0.0, 0.5, 0.5),
             'Y'   : (0.5, 0.0, 0.5),
             'Z'   : (0.5, 0.5, 0.0),}

  if ibrav_var=='ORCF3':
     band_path = 'gG-Y-T-Z-gG-X-A1-Y|X-A-Z|L-R'
     sorted_lat=np.sort([a,b,c])
     a=sorted_lat[0]
     b=sorted_lat[1]
     c=sorted_lat[2]
     eta =(1.0+(a/b)**2+(a/c)**2)/4.0
     zeta=(1.0+(a/b)**2+(a/c)**2)/4.0
     special_points = {'gG'  : (0.0, 0.0, 0.0),
             'A'   : (0.5, 0.5 + zeta, zeta),
             'A1'  : (0.5, 0.5-zeta, 1.0-zeta),
             'L'   : (0.5, 0.5, 0.5),
             'T'   : (1.0, 0.5, 0.5),
             'X'   : (0.0, eta, eta),
             'X1'  : (1.0, 1.0-eta, 1.0-eta),
             'Y'   : (0.5, 0.0, 0.5),
             'Z'   : (0.5, 0.5, 0.0)}

  if ibrav_var=='ORCC':
     band_path = 'gG-X-S-R-A-Z-gG-Y-X1-A1-T-Y|Z-T'

     zeta=(1.0+((a/b)**2))/4.0     

     special_points = {'gG'  : (  0.0, 0.0   , 0.0),
             'A'   : ( zeta, zeta  , 0.5),
             'A1'  : (-zeta, 1.0-zeta, 0.5),
             'R'   : (  0.0, 0.5   , 0.5),
             'S'   : (  0.0, 0.5   , 0.0),
             'T'   : ( -0.5, 0.5   , 0.5),
             'X'   : ( zeta, zeta  , 0.0),
             'X1'  : (-zeta, 1.0-zeta, 0.0),
             'Y'   : ( -0.5, 0.5   , 0.0),
             'Z'   : (  0.0, 0.0   , 0.5)}
     
  if ibrav_var=='ORCI':
     band_path = 'gG-X-L-T-W-R-X1-Z-gG-Y-S-W|L1-Y|Y1-Z'
     chi   = (1.0  + (a/c)**2)/(4.0)
     eta   = (1.0  + (b/c)**2)/(4.0)
     delta = (b**2 - a**2  )/(4.0*c**2)
     mu  = (b**2 + a**2  )/(4.0*c**2)
     special_points = {'gG'   : (0, 0, 0),
               'L'  : (-mu, mu, 0.5-delta),
               'L1'   : (mu, -mu, 0.5+delta),
               'L2'   : (0.5-delta, 0.5+delta, -mu),
               'R'  : (0.0, 0.5, 0.0),
               'S'  : (0.5, 0.0, 0.0),
               'T'  : (0.0, 0.0, 0.5),
               'W'  : (0.25,0.25,0.25),
               'X'  : (-chi, chi, chi),
               'X1'   : (chi, 1.0-chi, -chi),
               'Y'  : (eta, -eta, eta),
               'Y1'   : (1.0-eta, eta, -eta),
               'Z'  : (0.5, 0.5, -0.5)}

  if ibrav_var=='MCL':
     #abc->cab
     eta =  (1.0 - (b/a)*np.cos(np.pi-gamma))/(2.0*np.sin(np.pi-gamma)**2)
     nu =   0.5  - eta*(a/b)*np.cos(np.pi-gamma)
     band_path = 'gG-Y-H-C-E-M1-A-X-gG-Z-D-M|Z-A|D-Y|X-H1'
     special_points = {
               'gG'  : (0.0, 0.0  , 0.0  ),
               'A'   : (0.5, 0.5  , 0.0  ),
               'C'   : (0.0, 0.5  , 0.5  ),
               'D'   : (0.5, 0.0  , 0.5  ),
               'D1'  : (0.5, 0.0  ,-0.5  ),
               'E'   : (0.5, 0.5  , 0.5  ),
               'H'   : (0.0, eta  , 1.0-nu ),
               'H1'  : (0.0, 1.0-eta, nu   ),
               'H2'  : (0.0, eta  ,-nu   ),
               'M'   : (0.5, eta  , 1.0-nu ),
               'M1'  : (0.5, 1.0-eta, nu   ),
               'M2'  : (0.5, eta  ,-nu   ),
               'X'   : (0.0, 0.5  , 0.0  ),
               'Y'   : (0.0, 0.0  , 0.5  ),
               'Y1'  : (0.0, 0.0  ,-0.5  ),
               'Z'   : (0.5, 0.0  , 0.0  )}

     
   
  if ibrav_var=='TRI1A':     
    band_path = 'X-gG-Y|L-gG-Z|N-gG-M|R-gG' 
    special_points = {'gG'  : (0.0,0.0,0.0),
              'L'   : (0.5,0.5,0.0),
              'M'   : (0.0,0.5,0.5),
              'N'   : (0.5,0.0,0.5),
              'R'   : (0.5,0.5,0.5),
              'X'   : (0.5,0.0,0.0),
              'Y'   : (0.0,0.5,0.0),
              'Z'   : (0.0,0.0,0.5),}
    
  if ibrav_var=='TRI2A':    
    band_path = 'X-gG-Y|L-gG-Z|N-gG-M|R-gG'
    special_points = {'gG'  : (0.0,0.0,0.0),
              'L'   : (0.5,0.5,0.0),
              'M'   : (0.0,0.5,0.5),
              'N'   : (0.5,0.0,0.5),
              'R'   : (0.5,0.5,0.5),
              'X'   : (0.5,0.0,0.0),
              'Y'   : (0.0,0.5,0.0),
              'Z'   : (0.0,0.0,0.5),}
 
  if ibrav_var=='TRI1B':    
    band_path = "X-gG-Y|L-gG-Z|N-gG-M|R-gG"
    special_points = {'gG'  : ( 0.0, 0.0,0.0),
              'L'   : ( 0.5,-0.5,0.0),
              'M'   : ( 0.0, 0.0,0.5),
              'N'   : (-0.5,-0.5,0.5),
              'R'   : ( 0.0,-0.5,0.5),
              'X'   : ( 0.0,-0.5,0.0),
              'Y'   : ( 0.5, 0.0,0.0),
              'Z'   : (-0.5, 0.0,0.5),}

  if ibrav_var=='TRI2B':    
    band_path = 'X-gG-Y|L-gG-Z|N-gG-M|R-gG'
    special_points = {'gG'  : ( 0.0, 0.0,0.0),
              'L'   : ( 0.5,-0.5,0.0),
              'M'   : ( 0.0, 0.0,0.5),
              'N'   : (-0.5,-0.5,0.5),
              'R'   : ( 0.0,-0.5,0.5),

              'X'   : ( 0.0,-0.5,0.0),
              'Y'   : ( 0.5, 0.0,0.0),
              'Z'   : (-0.5, 0.0,0.5),}


  aflow_conv = np.identity(3)
  qe_conv  = np.identity(3)

  if ibrav==2:
    aflow_conv = np.asarray([[ 0.0, 1.0, 1.0],[ 1.0, 0.0, 1.0],[ 1.0, 1.0, 0.0]])/2.0   
    qe_conv  = np.asarray([[-1.0, 0.0, 1.0],[ 0.0, 1.0, 1.0],[-1.0, 1.0, 0.0]])/2.0
  if ibrav==3:
    aflow_conv = np.asarray([[-1.0, 1.0, 1.0],[ 1.0,-1.0, 1.0],[ 1.0, 1.0,-1.0]])/2.0   
    qe_conv  = np.asarray([[ 1.0, 1.0, 1.0],[-1.0, 1.0, 1.0],[-1.0,-1.0, 1.0]])/2.0   
  if ibrav==7:
    aflow_conv = np.asarray([[-1.0, 1.0, 1.0],[ 1.0,-1.0, 1.0],[ 1.0, 1.0,-1.0]])/2.0
    qe_conv  = np.asarray([[ 1.0,-1.0, 1.0],[ 1.0, 1.0, 1.0],[-1.0,-1.0, 1.0]])/2.0
  if ibrav==9:
    aflow_conv = np.asarray([[ 1.0,-1.0, 0.0],[ 1.0, 1.0, 0.0],[ 0.0, 0.0, 2.0]])/2.0
    qe_conv  = np.asarray([[ 1.0, 1.0, 0.0],[-1.0, 1.0, 0.0],[ 0.0, 0.0, 2.0]])/2.0
  if ibrav==10:
    aflow_conv = np.asarray([[ 0.0, 1.0, 1.0],[ 1.0, 0.0, 1.0],[ 1.0, 1.0, 0.0]])/2.0
    qe_conv  = np.asarray([[ 1.0, 0.0, 1.0],[ 1.0, 1.0, 0.0],[ 0.0, 1.0, 1.0]])/2.0  
  if ibrav==11:
    aflow_conv = np.asarray([[-1.0, 1.0, 1.0],[ 1.0,-1.0, 1.0],[ 1.0, 1.0,-1.0]])/2.0
    qe_conv  = np.asarray([[ 1.0, 1.0, 1.0],[-1.0, 1.0, 1.0],[-1.0,-1.0, 1.0]])/2.0
  if ibrav==12:
    aflow_conv = np.asarray([[ 0.0, 0.0, 1.0],[ 0.0, 1.0, 0.0],[ 1.0, 0.0, 0.0]])
    qe_conv  = np.asarray([[ 1.0, 0.0, 0.0],[ 0.0, 1.0, 0.0],[ 0.0, 0.0, 1.0]])
                   

  for k,v in special_points.items():
    first  = np.array(v).dot(np.linalg.inv(aflow_conv))
    if ibrav==9:
      second = qe_conv.T.dot(first)
    else:
      second = qe_conv.dot(first)
    special_points[k]=tuple((second).tolist())

  return special_points, band_path


def kpnts_interpolation_mesh ( data_controller ):
  '''
  Get path between HSP
  Arguments:
      nk (int): total number of points in path

  Returns:
      kpoints : array of arrays kx,ky,kz
      numK    : Total no. of k-points
  '''

  arrays,attr = data_controller.data_dicts()

  dk = 0.00001
  nk,alat,ibrav = attr['nk'],attr['alat'],attr['ibrav']
  a_vectors,b_vectors = arrays['a_vectors'],arrays['b_vectors']
  band_path,high_sym_points = attr['band_path'],arrays['high_sym_points']

  bp,hsp = (band_path,high_sym_points) if len(high_sym_points)!=0 else (None,None)

  points,_ = get_path(ibrav,alat,a_vectors,dk,b_vectors,bp,hsp)

  scaled_dk = dk*(points.shape[1]/nk)

  points,path_file = get_path(ibrav,alat,a_vectors,scaled_dk,b_vectors,bp,hsp)

  data_controller.write_kpnts_path('kpath_points.txt', path_file, points, b_vectors)

  arrays['kq'] = points


def get_path(ibrav,alat,cell,dk,b_vectors,band_path,special_points):

  def kdistance(hs, p1, p2):
    g = np.dot(hs.T, hs)
    p1, p2 = np.array(p1), np.array(p2)
    d = p1 - p2
    dist2 = np.dot(d.T, np.dot(g, d).T)
    return np.sqrt(dist2)

  def getSegments(path):
    segments = path.split('|')
    return segments

  def getPoints(pathSegment):
    pointsList = pathSegment.split('-')
    return pointsList
  def getNumPoints(path):
    list1 = getSegments(path)
    numPts = 0
    for index in (list1):
      numPts += len(getPoints(index))
    return numPts

  if ibrav==0 and band_path==None:
    sys.exit('IBRAV = 0 not permitted')
  if ibrav<0:
    print('Lattice type %s is not implemented') % ibrav
    logging.error('The ibrav value from QE has not yet been implemented')
    raise Exception

  totalK=0
  if band_path is None:
    special_points, band_path = _getHighSymPoints(ibrav,alat,cell)

  if type(special_points) is not dict:
    if rank == 0:
      print('ERROR: High symmetry points must be provided in the form of a dictionary with a label as the keys and a numpy array of points as the values\n')
    quit()

  hs = np.linalg.inv(cell)  # reciprocal lattice
  #hs = 2*np.pi*bcell
  segs = getSegments(band_path)

  kx = np.array([])
  ky = np.array([])
  kz = np.array([])

  path_file = ''

  for index in segs:

    a = getPoints(index) #gets the points in each segment of path separated by |
    point1 = None
    point2 = None

    for index2 in range(len(a)-1):
      try:
        point1 = a[index2]
        point2 = a[index2+1]

        p1 = special_points[point1]
        p2 = special_points[point2]
        
        newDK = (2.*np.pi/alat)*dk
        numK = int(np.ceil((kdistance(hs, p1, p2)/newDK)))

        totalK+=numK

        path_file+="%s %s\n"%(point1,numK)

        a0 = np.linspace(p1[0],p2[0],numK,endpoint=False)
        a1 = np.linspace(p1[1],p2[1],numK,endpoint=False)
        a2 = np.linspace(p1[2],p2[2],numK,endpoint=False)

        kx = np.concatenate((kx,a0))
        ky = np.concatenate((ky,a1))
        kz = np.concatenate((kz,a2))
        
      except Exception as e:
        print('Exception in kpnts_interpolation_mesh', e)

    """ last point """
    point1 = a[-1]
    p1 = special_points[point1]      
    kx = np.concatenate((kx,np.array([p1[0]])))
    ky = np.concatenate((ky,np.array([p1[1]])))
    kz = np.concatenate((kz,np.array([p1[2]])))
    totalK+=1

    path_file+="%s %s\n"%(a[-1],0)

  path_file+="\n"
  kpoints = np.array([kx,ky,kz])
    
  return kpoints,path_file



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def do_gradient ( data_controller ):
  import numpy as np
  from scipy import fftpack as FFT
  from .get_R_grid_fft import get_R_grid_fft
  from .communication import scatter_full

  arry,attr = data_controller.data_dicts()

  #----------------------
  # Compute the gradient of the k-space Hamiltonian
  #----------------------

  nktot = attr['nkpnts']
  snawf,nk1,nk2,nk3,nspin = arry['Hksp'].shape

  # fft grid in R shifted to have (0,0,0) in the center
  get_R_grid_fft(data_controller, nk1, nk2, nk3)

#  dHaux = np.empty((nk1*nk2*nk3,3), dtype=complex, order='C')
#  Haux = np.empty((nk1,nk2,nk3), dtype=complex, order='C')
  arry['dHksp'] = np.empty((snawf,nk1,nk2,nk3,3,nspin), dtype=complex, order='C')
  HRaux = np.empty((nk1*nk2*nk3))
  Dnm = scatter_full(np.reshape(arry['Dnm'],(attr['nawf']*attr['nawf'],3),order='C'), attr['npool'])
#  kdot = np.zeros((1,arry['R'].shape[0]), dtype=complex, order='C')
#  kdot = np.tensordot(arry['R'], -2.0j*np.pi*arry['kgrid'], axes=([1],[0]))
#  kdot = np.exp(kdot, kdot)
#  kdoti = np.zeros((1,arry['R'].shape[0]), dtype=complex, order='C')
#  kdoti = np.tensordot(arry['R'], 2.0j*np.pi*arry['kgrid'], axes=([1],[0]))
#  kdoti = np.exp(kdoti, kdoti)
  for ispin in range(nspin):
    for n in range(snawf):
      Haux = arry['Hksp'][n,:,:,:,ispin].copy()
      ########################################
      ### real space grid replaces k space ###
      ########################################
      if attr['use_cuda']:
        arry['Hksp'][n,:,:,:,ispin] = cuda_ifftn(arry['Hksp'][n,:,:,:,ispin])*1.0j*attr['alat']
      else:
        arry['Hksp'][n,:,:,:,ispin] = FFT.ifftn(arry['Hksp'][n,:,:,:,ispin])*1.0j*attr['alat']
        # HRaux = arry['Hksp'][n,:,:,:,ispin].reshape(attr['nk1']*attr['nk2']*attr['nk3'], order='C')
        # HRaux = np.tensordot(HRaux, kdoti, axes=([0],[0]))/(attr['nk1']*attr['nk2']*attr['nk3'])
        # arry['Hksp'][n,:,:,:,ispin] =  HRaux.reshape((nk1,nk2,nk3), order='C')*1.0j*attr['alat']

      # Compute R*H(R) + diagonal TB correction
      for l in range(3):
        arry['dHksp'][n,:,:,:,l,ispin] = FFT.fftn(arry['Rfft'][:,:,:,l]*arry['Hksp'][n,:,:,:,ispin]) +\
                                          1j*Haux[:,:,:]*Dnm[n,l]

      # HRaux = arry['Hksp'][n,:,:,:,ispin].reshape(attr['nk1']*attr['nk2']*attr['nk3'], order='C')
      # for l in range(3):
      #   dHaux = np.tensordot(HRaux, arry['R'][:,l]*kdot, axes=([0],[1]))
      #   arry['dHksp'][n,:,:,:,l,ispin] =  dHaux.reshape((nk1,nk2,nk3), order='C')

  



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
#F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli, Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np

from mpi4py import MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()

### Reformat
def build_Hks ( data_controller ):
  from scipy import linalg as spl

  arrays,attributes = data_controller.data_dicts()

  bnd = attributes['bnd']
  nawf = attributes['nawf']
  eta = attributes['shift']
  nspin = attributes['nspin']
  nkpnts = attributes['nkpnts']
  shift_type = attributes['shift_type']

  U = arrays['U'] 
  my_eigsmat = arrays['my_eigsmat']

  Hksaux = np.zeros((nawf,nawf,nkpnts,nspin), dtype=complex)
  Hks = np.zeros((nawf,nawf,nkpnts,nspin), dtype=complex)

  for ik in range(nkpnts):
    for ispin in range(nspin):
      my_eigs = my_eigsmat[:,ik,ispin]

      #Building the Hamiltonian matrix
      E = np.diag(my_eigs)
      UU = np.transpose(U[:,:,ik,ispin]) #transpose of U. Now the columns of UU are the eigenvector of length nawf
      norms = 1./np.sqrt(np.real(np.sum(np.conj(UU)*UU,axis=0)))
      UU[:,:nawf] = UU[:,:nawf]*norms[:nawf]

      # Choose only the eigenvalues that are below the energy shift
      bnd_ik = 0
      for n in range(bnd):
        if my_eigs[n] <= eta:
          bnd_ik += 1
      if bnd_ik == 0:
        if rank == 0:
          print('No Eigenvalues in the selected energy range')
        comm.Abort()
      ac = UU[:,:bnd_ik]  # filtering: bnd is defined by the projectabilities
      ee1 = E[:bnd_ik,:bnd_ik]
      if shift_type == 0:
        #option 1 (PRB 2013)
        Hksaux[:,:,ik,ispin] = ac.dot(ee1).dot(np.conj(ac).T) + eta*(np.identity(nawf)-ac.dot(np.conj(ac).T))

      elif shift_type == 1:
        #option 2 (PRB 2016)
        aux_p=spl.inv(np.dot(np.conj(ac).T,ac))
        Hksaux[:,:,ik,ispin] = ac.dot(ee1).dot(np.conj(ac).T) + eta*(np.identity(nawf)-ac.dot(aux_p).dot(np.conj(ac).T))

      elif shift_type == 2:
        # no shift
        Hksaux[:,:,ik,ispin] = ac.dot(ee1).dot(np.conj(ac).T)

      else:
        if rank == 0:
          print('\'shift_type\' Not Recognized')
        comm.Abort()

      # Enforce Hermiticity (just in case...)
      Hksaux[:,:,ik,ispin] = 0.5*(Hksaux[:,:,ik,ispin] + np.conj(Hksaux[:,:,ik,ispin].T))
      Hks = Hksaux
  return Hks


def do_build_pao_hamiltonian ( data_controller ):
  #------------------------------
  # Building the PAO Hamiltonian
  #------------------------------
  arry,attr = data_controller.data_dicts()

  arry['Hks'] = build_Hks(data_controller)

  ashape = (attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin'])

  if attr['expand_wedge']:
    from .pao_sym import open_grid_wrapper
    open_grid_wrapper(data_controller)

  attr['nkpnts'] = nkpnts = np.prod(ashape[2:5])

  # NOTE: Take care of non-orthogonality, if needed
  # Hks from projwfc is orthogonal. If non-orthogonality is required, we have to 
  # apply a basis change to Hks as Hks -> Sks^(1/2)+*Hks*Sks^(1/2)
  # acbn0 flag == 0 - makes H non orthogonal (original basis of the atomic pseudo-orbitals)
  # acbn0 flag == 1 - makes H orthogonal (rotated basis) 

  # if rank == 0 and attr['expand_wedge']:
  if rank == 0:
    from .do_Efermi import E_Fermi
    arry['Hks'] = np.reshape(arry['Hks'], ashape)

    # Shift the Fermi energy to zero
    tshape = (ashape[0],ashape[1],nkpnts,ashape[5])
   # Ef = E_Fermi(arry['Hks'].reshape(tshape), data_controller)
    Ef = 0
    dinds = np.diag_indices(attr['nawf'])
    arry['Hks'][dinds[0],dinds[1]] -= Ef

  if attr['acbn0']:
    import sys
    if rank == 0:
      from .do_non_ortho import do_non_ortho

      # This is needed for consistency of the ordering of the matrix elements
      # Important in ACBN0 file writing
      arry['Sks'] = np.transpose(arry['Sks'], (1,0,2))

      tshape = (ashape[0],ashape[1],nkpnts,ashape[5])
      arry['Hks'] = do_non_ortho(arry['Hks'].reshape(tshape), arry['Sks'])

      data_controller.write_Hk_acbn0()
    sys.exit(0)
    comm.Barrier()
    

def do_Hks_to_HRs ( data_controller ):
  from scipy import fftpack as FFT

  arry,attr = data_controller.data_dicts()

  #----------------------------------------------------------
  # Define the Hamiltonian and overlap matrix in real space:
  #   HRs and SRs (noinv and nosym = True in pw.x)
  #----------------------------------------------------------
  if rank == 0:
    # Original k grid to R grid
    arry['HRs'] = np.zeros_like(arry['Hks'])
    arry['HRs'] = FFT.ifftn(arry['Hks'], axes=[2,3,4])



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

from .do_atwfc_proj import *

# Function to calculate dipole matrix element from coefficients of wavefunction, 
# following the routine of epsilon.x
def calc_dipole(arry,attr, ik, ispin, b_vector):
  from scipy.io import FortranFile
  import os
  if attr['nspin'] == 1 or attr['nspin'] == 4:
    wfcfile = 'wfc{0}.dat'.format(ik+1)
  elif attr['nspin'] == 2 and ispin == 0:
    wfcfile = 'wfcdw{0}.dat'.format(ik+1)
  elif attr['nspin'] == 2 and ispin == 1:
    wfcfile = 'wfcup{0}.dat'.format(ik+1)
  else:
    print('no wfc file found')

  with FortranFile(os.path.join(attr['fpath'], wfcfile), 'r') as f:
    record = f.read_ints(np.int32)
    assert len(record) == 11, 'something wrong reading fortran binary file'

    ik_ = record[0]
    assert ik+1 == ik_, 'wrong k-point in wfc file???'

    # xk = np.frombuffer(record[1:7], np.float64)
    # ispin = record[7]
    # gamma_only = (record[8] != 0)
    scalef = np.frombuffer(record[9:], np.float64)[0]

    ngw, igwx, npol, nbnds = f.read_ints(np.int32)
    f.read_reals(np.float64).reshape(3,3,order='F')
    mill = f.read_ints(np.int32).reshape(3,igwx,order='F')
    mill = b_vector.T@mill + np.full((igwx,3),arry['kpnts'][ik]).T

    wfc = []
    for i in range(nbnds):
      wfc.append(f.read_reals(np.complex128))

  
  dipole_aux = np.zeros((3,nbnds,nbnds),dtype=np.complex128)
  for iband2 in range(nbnds):
    for iband1 in range(nbnds):
      if attr['dftSO']:
        dipole_aux[:,iband1,iband2] = (wfc[iband2][:igwx]*mill)@np.conjugate(wfc[iband1][:igwx]) 
        + (wfc[iband2][igwx:]*mill)@np.conjugate(wfc[iband1][igwx:])
      else:
        dipole_aux[:,iband1,iband2] = (wfc[iband2]*mill)@np.conjugate(wfc[iband1]) 
  return dipole_aux

# Function to calculate dipole matrix element from the eigenvector of the PAO Hamiltonian
# expanded in the real space of the atomic basis functions
def calc_dipole_internal(data_controller, ik, ispin):
  from .constants import RYTOEV

  arry, attr = data_controller.data_dicts()
  basis = arry['basis']
  gkspace = calc_gkspace(data_controller,ik,gamma_only=False)
  xk, igwx, mill, bg, _ = [gkspace[s] for s in ('xk', 'igwx', 'mill', 'bg', 'gamma_only')]
  atwfcgk = calc_atwfc_k(basis,gkspace)
  oatwfcgk = ortho_atwfc_k(atwfcgk) # these are the atomic orbitals on the G vector grid

  # build the full wavefunction with the coefficients v_k
  bnd = attr['bnd']
  wfc = []
  # for nb in range(attr['bnd']):
  for nb in range(bnd):
    wfc.append(np.tensordot(arry['v_k'][ik,:,nb,ispin],oatwfcgk,axes=(0,0)))

  # build k+G
  mill = arry['b_vectors'].T@mill + np.full((igwx,3),arry['kgrid'][:,ik]).T 

  nbnds = attr['nawf']
  dipole_aux = np.zeros((3,nbnds,nbnds),dtype=np.complex128)
  for iband2 in range(bnd):
    for iband1 in range(bnd):
      if attr['dftSO']:
        # check indexing with nbnds and bnd!!!!!
        dipole_aux[:,iband1,iband2] = (wfc[iband2][:igwx]*mill)@np.conjugate(wfc[iband1][:igwx]) 
        + (wfc[iband2][igwx:]*mill)@np.conjugate(wfc[iband1][igwx:])
      else:
        dipole_aux[:,iband1,iband2] = (wfc[iband2]*mill)@np.conjugate(wfc[iband1])
  return dipole_aux



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

class DataController:

  comm = rank = size = None

  data_arrays = data_attributes = None

  error_handler = report_exception = None

  def __init__ ( self, workpath, outputdir, inputfile, model, savedir, npool, smearing, acbn0, verbose, restart, dft ):
    '''
    Initialize the DataController
    Arguments:
        workpath (str): Path to the working directory
        outputdir (str): Name of the output directory (created in the working directory path)
        inputfile (str): (optional) Name of the xml inputfile
        npool (int): The number of pools to use. Increasing npool may reduce memory requirements.
        model (dict): A dictionary containing a model label and parameters
        savedir (str): QE .save directory
        acbn0 (bool): If True the Hamiltonian will be Orthogonalized after construction
        smearing (str): Smearing type (None, m-p, gauss)
        verbose (bool): False supresses debugging output
        restart (bool): True if the run is being restarted from a .json data dump.
    Returns:
        None
    '''
    from os import mkdir
    from mpi4py import MPI
    from os.path import join,exists
    from .ErrorHandler import ErrorHandler

    self.comm = MPI.COMM_WORLD
    self.rank = self.comm.Get_rank()
    self.size = self.comm.Get_size()

    if model is not None:
      if (inputfile is not None or savedir is not None) and self.rank == 0:
        print('\nWARNING: Model specified in addition to inputfile or savedir. Model will be used.')
    elif not restart and inputfile is None and savedir is None:
      if self.rank == 0:
        print('\nERROR: Must specify \'.save\' directory path, either in PAOFLOW constructor or in an inputfile.')
      quit()

    self.error_handler = ErrorHandler()
    self.report_exception = self.error_handler.report_exception

    if not restart and self.rank == 0:
      self.data_arrays = arrays = {}
      self.data_attributes = attr = {}

      # Set or update attributes
      attr['dft'] = dft
      attr['mpisize'] = self.size
      attr['savedir'] = savedir
      attr['verbose'] = verbose
      attr['workpath'] = workpath
      attr['acbn0'] = acbn0
      attr['inputfile'],attr['outputdir'] = inputfile,outputdir
      attr['opath'] = join(workpath, outputdir)
      if model is None:
        attr['fpath'] = join(workpath, (savedir if inputfile==None else inputfile))

      if inputfile == None:
        attr['temp'] = .025852
        attr['npool'],attr['smearing'] = npool,smearing

      # Create the output directory
      if not exists(attr['opath']):
        mkdir(attr['opath'])

      self.add_default_arrays()

      attr['abort_on_exception'] = True

      # Read inputfile, if it exsts
      if model is not None:
        from .defs.models import build_TB_model
        build_TB_model(self, model)
      else:
        try:
          if inputfile != None:
            if not exists(attr['fpath']):
              raise Exception('ERROR: Inputfile does not exist\n%s'%attr['fpath'])
            self.read_pao_inputfile()
          else:
            attr['do_spin_orbit'] = False
          if dft == 'QE':
            self.read_qe_output()
          else:
            self.read_vasp_output(symprec=1e-4)
        except Exception as e:
          print('\nERROR: Could not read xml data file')
          self.report_exception('Data Controller Initialization')
          raise e

    self.comm.Barrier()

    if not restart:
      # Broadcast Data
      try:
        self.data_arrays = self.comm.bcast(self.data_arrays, root=0)
        self.data_attributes = self.comm.bcast(self.data_attributes, root=0)
      except Exception as e:
        print('ERROR: MPI was unable to broadcast')
        self.report_exception('Initialization Broadcast')
        raise e


  def data_dicts ( self ):
    '''
    Get the data dictionaries

    Arguments:
        None

    Returns:
       (arrays, attributes): Tuple of two dictionaries, data_arrays and data_attributes 
    '''
    return(self.data_arrays, self.data_attributes)


  def print_data ( self ):
    '''
    Print the data dictionary keys

    Arguments:
        None

    Returns:
        None
    '''
    if self.rank == 0:
      print('\nData Attributes:')
      print(self.data_attributes.keys())
      print('\nData Arrays:')
      print(self.data_arrays.keys())
      print('\n')

  def build_arrays_adhoc_soc ( self ):
    '''
    Construc arrays needed for the adhoc_spin_orbit routine

    Arguments:
        None

    Returns:
        None
    '''
      
    arry,attr = self.data_dicts()
    orb=[]
    naw=[]

    for i in range (len(arry['atoms'])):
      if (arry['shells'][arry['atoms'][i]] == [0]):
        naw.append(1)
        orb.append('s')
      if (arry['shells'][arry['atoms'][i]] == [0,1]):
        naw.append(4)
        orb.append('sp')
      if (arry['shells'][arry['atoms'][i]] == [0,1,2]):
        naw.append(9)
        orb.append('spd')
      if (arry['shells'][arry['atoms'][i]] == [1,0]):
        naw.append(4)
        orb.append('ps')
      if (arry['shells'][arry['atoms'][i]] == [0,0,1,2]):
        naw.append(10)
        orb.append('sspd')
      if (arry['shells'][arry['atoms'][i]] == [0,0,1]):
        naw.append(5)
        orb.append('ssp')
      if (arry['shells'][arry['atoms'][i]] == [0,0,1,1,2]):
        naw.append(13)
        orb.append('ssppd')

    arry['orb_pseudo'] = orb
    arry['naw'] = naw

  def add_default_arrays ( self ):
    import numpy as np

    # Band Path
    self.data_attributes['band_path'] = None
    self.data_arrays['high_sym_points'] = {}

    # Electric Field
    self.data_arrays['Efield'] = np.zeros(3, dtype=float)
    # Magnetic Field
    self.data_arrays['Bfield'] = np.zeros(3, dtype=float)
    # Hubbard Parameters
    self.data_arrays['HubbardU'] = np.zeros(1, dtype=float)

    # Tensor components
    # Dielectric function
    self.data_arrays['d_tensor'] = np.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2],[2,0],[2,1],[2,2]])
    # Boltzmann transport
    self.data_arrays['t_tensor'] = np.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2],[2,0],[2,1],[2,2]])
    # Berry curvature
    self.data_arrays['a_tensor'] = np.array([[0,0],[0,1],[0,2],[1,0],[1,1],[1,2],[2,0],[2,1],[2,2]])
    # Spin Berry curvature
    self.data_arrays['s_tensor'] = np.array([[0,0,0],[0,1,0],[0,2,0],[1,0,0],[1,1,0],[1,2,0],[2,0,0],[2,1,0],[2,2,0], \
                                             [0,0,1],[0,1,1],[0,2,1],[1,0,1],[1,1,1],[1,2,1],[2,0,1],[2,1,1],[2,2,1], \
                                             [0,0,2],[0,1,2],[0,2,2],[1,0,2],[1,1,2],[1,2,2],[2,0,2],[2,1,2],[2,2,2]])


  def read_pao_inputfile ( self ):
    from .defs.read_inputfile_xml_parse import read_inputfile_xml
    read_inputfile_xml(self.data_attributes['workpath'], self.data_attributes['inputfile'], self)

  def read_qe_output ( self ):
    from os.path import exists
    fpath = self.data_attributes['fpath']

    if exists(fpath+'/data-file-schema.xml'):
      from .defs.read_QE_xml import parse_qe_data_file_schema
      parse_qe_data_file_schema(self, fpath+'/data-file-schema.xml')
    elif exists(fpath+'/data-file.xml'):
      from .defs.read_QE_xml import parse_qe_data_file
      parse_qe_data_file(self, fpath, 'data-file.xml')
    else:
      raise Exception('data-file.xml or data-file-schema.xml were not found.\n')

  def read_vasp_output ( self, symprec ):
    from os.path import exists
    fpath = self.data_attributes['fpath']

    if exists(fpath+'/vasprun.xml'):
      from .defs.read_VASP import parse_vasprun_data
      parse_vasprun_data(self, fpath+'/vasprun.xml', symprec)
    else:
      raise Exception('vasprun.xml was not found.\n')


  def write_file_row_col ( self, fname, col1, col2 ):
    '''
    Write a file with 2 columns

    Arguments:
        fname (str): Name of the file (written to outputdir)
        col1 (ndarray): 1D array of values for the rightmost column
        col2 (ndarray): 1D array of values for the leftmost column

    Returns:
        None
    '''
    if self.rank == 0:
      from os.path import join
      if len(col1) != len(col2):
        print('ERROR: Cannot write file: %s'%fname)
        print('Data does not have the same shape')
        self.comm.Abort()

      attr = self.data_attributes

      with open(join(attr['opath'],fname), 'w') as f:
        for i in range(len(col1)):
          f.write('%.5f %.15e\n'%(col1[i],col2[i]))
    self.comm.Barrier()


  def write_bxsf ( self, fname, bands, nbnd, indices=None ):
    '''
    Write a file in the bxsf format

    Arguments:
        fname (str): Name of the file (written to outputdir)
        bands (ndarray): The data array of values to be written
        nbnd (int): Number of values from array to write

    Returns:
        None
    '''
    attr = self.data_attributes

    if self.rank == 0:
      from .defs.write2bxsf import write2bxsf
      from .defs.write2bxsf4skeaf import write2bxsf4skeaf

      write2bxsf(self, fname, bands, nbnd, indices, attr['fermi_up'], attr['fermi_dw'])
      write2bxsf4skeaf(self, fname, bands, nbnd, indices, attr['fermi_up'], attr['fermi_dw'])


  def write_bands ( self, fname, bands ):
    '''
    Write electronic band structure file

    Arguments:
        fname (str): Name of the file (written to outputdir)
        bands (ndarray): Band data array (shape: (nk,nbnd,nspin))

    Returns:
        None
    '''
    if self.rank == 0:
      from os.path import join

      arry,attr = self.data_dicts()
      nspin,nkpi = attr['nspin'],arry['kq'].shape[1]

      for ispin in range(nspin):
        with open(join(attr['opath'],fname+'_'+str(ispin)+'.dat'), 'w') as f:
          for ik in range(nkpi):
            f.write(' '.join(['%6d'%ik]+['% 14.8f'%j for j in bands[ik,:,ispin]])+'\n') 
    self.comm.Barrier()


  def write_kpnts_path ( self, fname, path, kpnts, b_vectors ):
    '''
    Write the k-path through the BZ

    Arguments:
        fname (str): Name of the file (written to outputdir)
        kpnts (ndarray): k-point path (shape: (3,nk))

    Returns:
        None
    '''
    if self.rank == 0:
      from os.path import join

      kpnts= b_vectors.T.dot(kpnts)
      with open(join(self.data_attributes['opath'],fname), 'w') as f:
        f.write(path)
        f.write(''.join(['%s %s %s\n'%(kpnts[0,i],kpnts[1,i],kpnts[2,i]) for i in range(kpnts.shape[1])]))
    self.comm.Barrier()


#  def write_xsf ( self, data, fname='rs_plot.xsf' ):
#    import numpy as np
#    
#    if self.rank == 0:
#      from os.path import join
#
#      fpath = self.data_attributes['opath']
#      if data.dtype == complex:
#        data = np.abs(data)
#      atoms = self.data_arrays['tau']
#      species = self.data_arrays['atoms']
#      a_vecs = self.data_attributes['alat'] * self.data_arrays['a_vectors'] * 0.529177
#
#      with open(join(fpath,fname), 'w') as f:
#        f.write('CRYSTAL\nPRIMVEC\n')
#        for i in range(3):
#          f.write(' %.14f %.14f %.14f\n'%tuple(a_vecs[i]))
#        f.write('PRIMCOORD\n%d 1\n'%atoms.shape[0])
#        for i in range(atoms.shape[0]):
#          tup = (species[i],) + tuple(atoms[i])
#          f.write(' %s %20.14f %20.14f %20.14f\n'%tup)
#
#        f.write('BEGIN_BLOCK_DATAGRID_3D\n data\n BEGIN_DATAGRID_3Dgrid#1\n')
#
#        shape = data.shape
#        f.write(' %d %d %d\n'%shape)
#        f.write(' 0 0 0\n') # Origin
#        for i in range(3):
#          f.write(' %f %f %f\n'%tuple(a_vecs[i] * (shape[i]-1) / shape[i]))
#
#        for k in range(shape[2]):
#          for j in range(shape[1]):
#            f.write(' ' + ' '.join(['%f' % d for d in data[:, j, k]]) + '\n')
#          f.write('\n')
#
#        f.write('END_DATAGRID_3D\nEND_BLOCK_DATAGRID_3D\n')
#
#    self.comm.Barrier()


  def write_Hk_acbn0 ( self ):
    #----------------------
    # write to file Hks,Sks,kpnts,kpnts_wght
    #---------------------- 
    import numpy as np
    import os,sys

    if self.rank == 0:

      arry,attr    = self.data_dicts()
      nspin        = attr['nspin']
      inputpath    = attr['opath']
      write_binary = attr['write_binary']
      nawf         = attr['nawf']
      acbn0    = attr['acbn0']
      kpnts_wght   = arry['kpnts_wght']
      kpnts        = arry['kpnts']
      Hks          = arry['Hks']
      Sks          = arry['Sks']
      nkpnts       = kpnts.shape[0]

      Skss = Sks.shape
      if len(Skss) > 3:
        if len(Skss) == 5:
          Sks = Sks.reshape((Skss[0],Skss[1],Skss[2]*Skss[3]*Skss[4]))
        else:
          Sks = Sks.reshape((Skss[0],Skss[1],Skss[2]*Skss[3]*Skss[4],Skss[5]))

      if write_binary:# or whatever you want to call it            
        if nspin==1:#postfix .npy just to make it clear what they are
          np.save(os.path.join(inputpath,'kham.npy'),np.ravel(Hks[...,0],order='C'))
        if nspin==2:
          np.save(os.path.join(inputpath,'kham_up.npy'),np.ravel(Hks[...,0],order='C'))
          np.save(os.path.join(inputpath,'kham_dn.npy'),np.ravel(Hks[...,1],order='C'))
        if acbn0:
          Sks = Sks[:Sks.shape[1],:,:]
          np.save(os.path.join(inputpath,'kovp.npy'),np.ravel(Sks,order='C'))
      else:
        if nspin == 1:
          f=open(os.path.join(inputpath,'kham.txt'),'w')
          for ik in range(nkpnts):
            for i in range(nawf):
              for j in range(nawf):
                f.write('%20.13f %20.13f \n' %(np.real(Hks[i,j,ik,0]),np.imag(Hks[i,j,ik,0])))
          f.close()
        elif nspin == 2:
          f=open(os.path.join(inputpath,'kham_up.txt'),'w')
          for ik in range(nkpnts):
            for i in range(nawf):
              for j in range(nawf):
                f.write('%20.13f %20.13f \n' %(np.real(Hks[i,j,ik,0]),np.imag(Hks[i,j,ik,0])))
          f.close()
          f=open(os.path.join(inputpath,'kham_down.txt'),'w')
          for ik in range(nkpnts):
            for i in range(nawf):
              for j in range(nawf):
                f.write('%20.13f %20.13f \n' %(np.real(Hks[i,j,ik,1]),np.imag(Hks[i,j,ik,1])))
          f.close()
      if acbn0:
          f=open(os.path.join(inputpath,'kovp.txt'),'w')
          for ik in range(nkpnts):
            for i in range(nawf):
              for j in range(nawf):
                f.write('%20.13f %20.13f \n' %(np.real(Sks[i,j,ik]),np.imag(Sks[i,j,ik])))
          f.close()
      f=open(os.path.join(inputpath,'k.txt'),'w')
      for ik in range(nkpnts):
        f.write('%20.13f %20.13f %20.13f \n' %(kpnts[ik,0],kpnts[ik,1],kpnts[ik,2]))
      f.close()
      f=open(os.path.join(inputpath,'wk.txt'),'w')
      for ik in range(nkpnts):
        f.write('%20.13f \n' %(kpnts_wght[ik]))
      f.close()

      print('H(k),S(k),k,wk written to file')


  def write_z2pack ( self, fname ):
    '''
    Write HRs in the Z2 Pack format

    Arguments:
        fname (str): Name of the file (written to outputdir)

    Returns:
        None
    '''
    try:
      if self.rank == 0:
        import numpy as np
        from os.path import join
        from .defs.zero_pad import zero_pad

        arry,attr = self.data_dicts()

        HRS=np.fft.ifftn(arry["Hks"],axes=(2,3,4))

        nawf,_,nk1,nk2,nk3,nspin=HRS.shape
        # how to pad HR to make sure it's odd for z2pack
        if nk1%2:
          pad1=0
        else: pad1=1
        if nk2%2:
          pad2=0
        else: pad2=1
        if nk3%2:
          pad3=0
        else: pad3=1


        HRS_interp=np.zeros((nawf,nawf,nk1+pad1,nk2+pad2,nk3+pad3,nspin),dtype=complex)
        for n in range(nawf):
          for m in range(nawf):
            for ispin in range(nspin):
              HRS_interp[n,m,:,:,:,ispin] = zero_pad(HRS[n,m,:,:,:,ispin],nk1,nk2,nk3,pad1,pad2,pad3)

        nk1+=pad1
        nk2+=pad2
        nk3+=pad3
        HRS=HRS_interp
        HRS_interp=None

        with open(join(attr['opath'],fname), 'w') as f:#'z2pack_hamiltonian.dat','w')
  
          nkpts = nk1*nk2*nk3
          f.write("PAOFLOW Generated \n")
          f.write('%5d \n'%nawf)
  
          f.write('%5d \n'%(nk1*nk2*nk3))

          nl = 15 # z2pack read the weights in lines of 15 items
  
          nlines = nkpts//nl # number of lines
          nlast = nkpts%nl   # number of items of laste line if needed
  
          for j in range(nlines):
            f.write("1 "*nl)
            f.write("\n")
          f.write("1 "*nlast)
          f.write("\n")
  
  #### Can be condensed
          for i in range(nk1):
            for j in range(nk2):
              for k in range(nk3):
                n = k + j*nk3 + i*nk2*nk3
                Rx = float(i)/float(nk1)
                Ry = float(j)/float(nk2)
                Rz = float(k)/float(nk3)
                if Rx >= 0.5: Rx=Rx-1.0
                if Ry >= 0.5: Ry=Ry-1.0
                if Rz >= 0.5: Rz=Rz-1.0
                Rx -= int(Rx)
                Ry -= int(Ry)
                Rz -= int(Rz)
                # the minus sign in Rx*nk1 is due to the Fourier transformation (Ri-Rj)
                ix=-round(Rx*nk1,0)
                iy=-round(Ry*nk2,0)
                iz=-round(Rz*nk3,0)
                for m in range(nawf):
                  for l in range(nawf):
                    # l+1,m+1 just to start from 1 not zero

                    f.write('%3d %3d %3d %5d %5d %28.14f %28.14f\n'%(ix,iy,iz,l+1,m+1,
                                                               HRS[l,m,i,j,k,0].real,
                                                               HRS[l,m,i,j,k,0].imag,))
                                                               
    except Exception as e:
      self.report_exception('z2_pack')
      if self.data_attributes['abort_on_exception']:
        raise e

    self.comm.Barrier()



  def broadcast_single_array ( self, key, dtype=complex, root=0 ):
    '''
    Broadcast array from 'data_arrays' with 'key' from 'root' to all other ranks

    Arguments:
        key (str): The key for the array to broadcast (key must exist in dictionary 'data_arrays')
        dtype (dtype): The data type of the array to broadcast
        root (int): The rank which is the source of the broadcasted array

    Returns:
        None
    '''
    import numpy as np
    ashape = self.comm.bcast((None if self.rank!=root else self.data_arrays[key].shape), root=root)
    if self.rank != root:
      self.data_arrays[key] = np.zeros(ashape, dtype=dtype, order='C')
    self.comm.Bcast(np.ascontiguousarray(self.data_arrays[key]), root=root)

  def broadcast_attribute ( self, key, root=0 ):
    '''
    Broadcast attribute from 'data_attributes' with 'key' from 'root' to all other ranks

    Arguments:
        key (str): The key for the attribute to broadcast (key must exist in dictionary 'data_attributes')
        root (int): The rank which is the source of the broadcasted attribut

    Returns:
        None
    '''
    attr = self.data_attributes[key] if self.rank==root else None
    self.data_attributes[key] = self.comm.bcast(attr, root=root)

### This section is under construction
### Only 'broadcast_single_array' should be used
  def broadcast_data_attributes ( self ):
    if self.rank == 0:
      for i in range(1,self.size):
        self.comm.send(self.data_attributes, dest=i)
    else:
      self.data_attributes = self.comm.recv(source=0)

  def broadcast_data_arrays ( self ):
    if self.rank == 0:
      for i in range(1,self.size):
        self.comm.send(self.data_arrays, dest=i)
    else:
      self.data_arrays = self.comm.recv(source=0)

  def scatter_data_array ( self, key ):
    from .defs.communication import scatter_array
    self.data_arrays[key] = scatter_array(self.data_arrays[key])

  def gather_data_array ( self, key ):
    import numpy as np
    from .defs.communication import gather_array
    arr = self.data_arrays[key]
    aux = None
    size1 = arr.shape[0]

    if self.rank == 0:
      for i in range(1,self.size):
        size1 += comm.recv(source=i)
    else:
      comm.send(arr.shape[0], dest=0)

    if self.rank == 0:
      if len(arr.shape) > 1:
        aux = np.ndarray(shape=((size1,)+arr.shape[1:]), dtype=arr.dtype)
      else:
        aux = np.ndarray(shape=(size1), dtype=arr.dtype)

    gather_array(aux, arr)

    if self.rank == 0:
      self.data_arrays[key] = aux
    else:
      self.data_arrays[key] = None
